"""
Wrapper over the NetworkManager client.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
# pylint: disable=duplicate-code

from concurrent.futures import Future
from threading import Thread, Lock
from typing import Optional, List

from packaging.version import Version

import gi
gi.require_version("NM", "1.0")
from gi.repository import NM, GLib, Gio, GObject  # pylint: disable=C0413 # noqa: E402

from proton.vpn import logging  # noqa: E402 pylint: disable=wrong-import-position

logger = logging.getLogger(__name__)


class GatewayNotFoundError(Exception):
    """Error raised when trying to obtain the gateway from a device that
    doesn't have one set."""


def _create_future():
    """Creates a future and sets its internal state as running."""
    future = Future()
    future.set_running_or_notify_cancel()
    return future


class NMClient:
    """
    Wrapper over the NetworkManager client.
    It also starts the GLib main loop used by the NetworkManager client.
    """
    _lock = Lock()
    _main_context = None
    _nm_client = None

    @classmethod
    def initialize_nm_client_singleton(cls):
        """
        Initializes the NetworkManager client singleton.

        If the singleton was initialized, this method will do nothing. However,
        if the singleton wasn't initialized it will initialize it, starting
        a new GLib MainLoop.

        A double-checked lock is used to avoid the possibility of multiple
        threads concurrently creating multiple instances of the NM client
        (with their own main loops).
        """
        if cls._nm_client:
            return

        with cls._lock:
            if not cls._nm_client:
                cls._initialize_nm_client_singleton()

    @classmethod
    def _initialize_nm_client_singleton(cls):
        cls._main_context = GLib.MainContext()

        # Setting daemon=True when creating the thread makes that this thread
        # exits abruptly when the python process exits. It would be better to
        # exit the thread running the main loop calling self._main_loop.quit().
        Thread(target=cls._run_glib_loop, daemon=True).start()

        def _init_nm_client():
            # It's important the NM.Client instance is created in the thread
            # running the GLib event loop so that then that's the thread used
            # for all GLib asynchronous operations.
            return NM.Client.new(cancellable=None)

        cls._nm_client = cls._run_on_glib_loop_thread(_init_nm_client).result()

    @classmethod
    def _run_glib_loop(cls):
        main_loop = GLib.MainLoop(cls._main_context)
        cls._main_context.push_thread_default()
        main_loop.run()

    @classmethod
    def _assert_running_on_glib_loop_thread(cls):
        """
        This method asserts that the thread running it is the one i:terating
        GLib's main loop.

        It's useful to call this method at the beginning of any code block
        that's supposed to run in GLib's main loop, to avoid hard-to-debug
        issues.

        For more info:
        https://developer.gnome.org/documentation/tutorials/main-contexts.html#checking-threading
        """
        if not cls._main_context.is_owner():
            raise RuntimeError("Code being run outside GLib's main loop.")

    @classmethod
    def _run_on_glib_loop_thread(cls, function, *args, **kwargs) -> Future:
        future = _create_future()

        def wrapper():
            cls._assert_running_on_glib_loop_thread()
            try:
                future.set_result(function(*args, **kwargs))
            except BaseException as exc:  # pylint: disable=broad-except
                future.set_exception(exc)

        cls._main_context.invoke_full(priority=GLib.PRIORITY_DEFAULT, function=wrapper)

        return future

    def __init__(self):
        self.initialize_nm_client_singleton()
        self._signal_handlers = []

    def add_connection_async(
            self, connection: NM.Connection, save_to_disk: bool = False
    ) -> Future:
        """
        Adds a new connection asynchronously.
        https://lazka.github.io/pgi-docs/#NM-1.0/classes/Client.html#NM.Client.add_connection_async
        :param connection: connection to be added.
        :return: a Future to keep track of completion.
        """
        future_conn_activated = _create_future()

        def _on_interface_state_changed(_device, new_state, _old_state, _reason):
            """
            Monitors kill switch interface state changes and resolves
            the future as soon as the interface reaches the activated state
            """
            logger.debug(
                f"{connection.get_interface_name()} interface state changed "
                f"to {NM.DeviceState(new_state).value_name}"
            )
            if (
                    NM.DeviceState(new_state) == NM.DeviceState.ACTIVATED
                    and not future_conn_activated.done()
            ):
                future_conn_activated.set_result(None)

        def _on_interface_added(_nm_client, device):
            """
            Monitors interface creation. As soon as the kill switch interface
            is created it sets up the call back to monitor interface state changes.
            """
            logger.debug(
                f"{device.get_iface()} interface added in state {device.get_state().value_name}"
            )
            if not device.get_iface() == connection.get_interface_name():
                return

            handler_id = device.connect("state-changed", _on_interface_state_changed)
            future_conn_activated.add_done_callback(
                lambda f: self._run_on_glib_loop_thread(
                    GObject.signal_handler_disconnect, device, handler_id
                ).result()
            )

        def _on_connection_added(nm_client, res, _user_data):
            try:
                # Make sure exceptions creating the connection are passed to the future.
                nm_client.add_connection_finish(res)
            except Exception as exc:  # pylint: disable=broad-except
                future_conn_activated.set_exception(
                    RuntimeError(
                        f"Error adding KS connection: {exc}"
                    ).with_traceback(exc.__traceback__)
                )
                return

        def _add_connection_async():
            # Set up interface connection monitoring, which resolves the future
            # once the kill switch is active.
            handler_id = self._nm_client.connect("device-added", _on_interface_added)
            future_conn_activated.add_done_callback(
                lambda f: self._run_on_glib_loop_thread(
                    GObject.signal_handler_disconnect, self._nm_client, handler_id
                ).result()
            )

            # Add kill switch connection asynchronously.
            self._nm_client.add_connection_async(
                connection=connection,
                save_to_disk=save_to_disk,
                cancellable=None,
                callback=_on_connection_added,
                user_data=None
            )

        self._run_on_glib_loop_thread(_add_connection_async).result()

        return future_conn_activated

    def get_physical_devices(self) -> List[NM.Device]:
        """Returns all the active ethernet/wifi devices."""
        return [
            device for device in self._nm_client.get_devices() if (
                device.get_device_type() in (NM.DeviceType.ETHERNET, NM.DeviceType.WIFI)
                and device.get_state() is NM.DeviceState.ACTIVATED
                and device.get_active_connection()  # Maybe this is redundant.
            )
        ]

    def _is_ethernet_or_wifi_connection(self, active_connection: NM.ActiveConnection):
        return active_connection.props.type in (
            NM.SETTING_WIRED_SETTING_NAME, NM.SETTING_WIRELESS_SETTING_NAME
        )

    def get_ethernet_and_wifi_connections(self) -> List[NM.ActiveConnection]:
        """Returns the list of ethernet and wifi connections."""
        return [
            ac for ac in self._nm_client.get_active_connections()
            if self._is_ethernet_or_wifi_connection(ac)
        ]

    def is_monitoring_network_config_changes(self) -> bool:
        """
        Returns whether monitoring of network configuration changes is already
        enabled or not.
        """
        return bool(self._signal_handlers)

    def start_monitoring_network_config_changes(self, callback):
        """Starts monitoring network configuration changes."""
        if self.is_monitoring_network_config_changes():
            raise RuntimeError("Network config changes are already being monitored.")

        def on_active_connection_changed(active_connection, changed_field):
            if (
                    changed_field.name == NM.DEVICE_IP4_CONFIG
                    and active_connection.get_ip4_config()
                    and active_connection.get_ip4_config().get_gateway()
            ):
                callback(active_connection)

        # Monitor current active (wifi/ethernet) connection changes.
        connections = self.get_ethernet_and_wifi_connections()
        for con in connections:
            handler_id = con.connect("notify", on_active_connection_changed)
            self._signal_handlers.append((con, handler_id))

        def on_active_connection_added(_nm_client, active_connection):
            if self._is_ethernet_or_wifi_connection(active_connection):
                # If the new active connection is a wifi/ethernet connection then
                # start monitoring its config changes.
                handler_id = active_connection.connect("notify", on_active_connection_changed)
                self._signal_handlers.append((active_connection, handler_id))

        # Monitor new active connections.
        handler_id = self._nm_client.connect("active-connection-added", on_active_connection_added)
        self._signal_handlers.append((self._nm_client, handler_id))

    def stop_monitoring_network_config_changes(self):
        """Stops monitoring network configuration changes."""
        for instance, handler_id in self._signal_handlers:
            instance.disconnect(handler_id)

        self._signal_handlers.clear()

    @classmethod
    def _add_ipv4_route(
            cls, active_connection: NM.ActiveConnection,
            server_ip: str, gateway: str
    ):
        cls._assert_running_on_glib_loop_thread()

        connection = active_connection.get_connection()
        config = connection.get_setting_ip4_config()

        config.add_route(
            NM.IPRoute.new(
                family=GLib.SYSDEF_AF_INET,
                dest=server_ip,
                prefix=32,
                next_hop=gateway,
                metric=-1  # -1 just means that the default metric is applied.
            )
        )

    @classmethod
    def _remove_ipv4_routes(
            cls, active_connection: NM.ActiveConnection,
            server_ip: str
    ):
        cls._assert_running_on_glib_loop_thread()

        connection = active_connection.get_connection()
        config = connection.get_setting_ip4_config()

        routes_to_remove = []
        for i in range(config.get_num_routes()):
            route = config.get_route(i)
            if route.get_dest() == server_ip and route.get_prefix() == 32:
                routes_to_remove.append(route)

        for route in routes_to_remove:
            config.remove_route_by_value(route)

    @classmethod
    def _apply_connection_async(
            cls, active_connection: NM.ActiveConnection, future: Future
    ):
        cls._assert_running_on_glib_loop_thread()

        device = active_connection.get_devices()[0]

        def on_device_reapplied(device, result, _data=None):
            try:
                device.reapply_finish(result)
                future.set_result(None)
            except Exception as exc:  # pylint: disable=broad-except
                future.set_exception(exc)

        def on_connection_commited(connection, result, _data=None):
            try:
                connection.commit_changes_finish(result)
                device.reapply_async(
                    # Not sure if it's ok to always pass version_id and flags set to 0.
                    connection, version_id=0, flags=0,
                    cancellable=None, callback=on_device_reapplied
                )
            except Exception as exc:  # pylint: disable=broad-except
                future.set_exception(exc)

        # By not saving the changes to disk, the route is gone after a restart.
        active_connection.get_connection().commit_changes_async(
            save_to_disk=False, cancellable=None, callback=on_connection_commited
        )

    @classmethod
    def add_route_to_device(
            cls, device: NM.Device, new_server_ip: str, old_server_ip: Optional[str] = None
    ) -> Future:
        """
        Adds a route to the device to reach the new server ip via the gateway configured
        on the device.

        :param device: the device to apply the route to.
        :param new_server_ip: the IP to apply the route for.
        :param old_server_ip: if specified, routes for this IP will be removed before
          adding the new one.
        """
        route_added_future = _create_future()
        active_connection = device.get_active_connection()

        def _add_ipv4_route():
            try:
                if old_server_ip:
                    cls._remove_ipv4_routes(active_connection, old_server_ip)

                # Remove any existing routes to the new server that may have been left over.
                cls._remove_ipv4_routes(active_connection, new_server_ip)

                gateway = active_connection.get_ip4_config().get_gateway()
                if not gateway:
                    raise GatewayNotFoundError(
                        "Gateway not found on interface "
                        f"'{device.get_iface()}'"
                    )

                cls._add_ipv4_route(active_connection, new_server_ip, gateway)

                cls._apply_connection_async(active_connection, route_added_future)
            except Exception as exc:  # pylint: disable=broad-except
                route_added_future.set_exception(exc)

        cls._run_on_glib_loop_thread(_add_ipv4_route)

        return route_added_future

    @classmethod
    def remove_route_from_device(cls, device: NM.Device, server_ip: str) -> Future:
        """
        Removes all the routes pointing to the specified server IP
        for the specified device.
        """
        route_removed_future = _create_future()
        active_connection = device.get_active_connection()

        def _remove_ipv4_routes():
            try:
                cls._remove_ipv4_routes(active_connection, server_ip)
                cls._apply_connection_async(active_connection, route_removed_future)
            except Exception as exc:  # pylint: disable=broad-except
                route_removed_future.set_exception(exc)

        cls._run_on_glib_loop_thread(_remove_ipv4_routes)

        return route_removed_future

    def remove_connection_async(
            self, connection: NM.RemoteConnection
    ) -> Future:
        """
        Removes the specified connection asynchronously.
        https://lazka.github.io/pgi-docs/#NM-1.0/classes/RemoteConnection.html#NM.RemoteConnection.delete_async
        :param connection: connection to be removed.
        :return: a Future to keep track of completion.
        """
        future_interface_removed = _create_future()

        def _on_connection_removed(connection, result, _user_data):
            try:
                connection.delete_finish(result)
            except Exception as exc:  # pylint: disable=broad-except
                future_interface_removed.set_exception(
                    RuntimeError(
                        f"Error removing KS connection: {exc}"
                    ).with_traceback(exc.__traceback__)
                )

        def _on_interface_removed(_nm_client, device):
            logger.debug(
                f"{device.get_iface()} was removed."
            )
            if device.get_iface() == connection.get_interface_name():
                future_interface_removed.set_result(None)

        def _remove_connection_async():
            handler_id = self._nm_client.connect("device-removed", _on_interface_removed)
            future_interface_removed.add_done_callback(
                lambda f: self._run_on_glib_loop_thread(
                    GObject.signal_handler_disconnect, self._nm_client, handler_id
                ).result()
            )

            connection.delete_async(
                None,
                _on_connection_removed,
                None
            )

        self._run_on_glib_loop_thread(_remove_connection_async).result()

        return future_interface_removed

    def get_active_connection(self, conn_id: str) -> Optional[NM.ActiveConnection]:
        """
        Returns the specified active connection, if existing.
        :param conn_id: ID of the active connection.
        :return: the active connection if it was found. Otherwise, None.
        """
        def _get_active_connection():
            active_connections = self._nm_client.get_active_connections()

            for connection in active_connections:
                if connection.get_id() == conn_id:
                    return connection

            return None

        return self._run_on_glib_loop_thread(_get_active_connection).result()

    def get_connection(self, conn_id: str) -> Optional[NM.RemoteConnection]:
        """
        Returns the specified connection, if existing.
        :param conn_id: ID of the connection.
        :return: the connection if it was found. Otherwise, None.
        """
        return self._run_on_glib_loop_thread(
            self._nm_client.get_connection_by_id, conn_id
        ).result()

    def get_nm_running(self) -> bool:
        """Returns if NetworkManager daemon is running or not."""
        return self._run_on_glib_loop_thread(
            self._nm_client.get_nm_running
        ).result()

    def connectivity_check_get_enabled(self) -> bool:
        """Returns if connectivity check is enabled or not."""
        return self._run_on_glib_loop_thread(
            self._nm_client.connectivity_check_get_enabled
        ).result()

    def disable_connectivity_check(self) -> Future:
        """Since `connectivity_check_set_enabled` has been deprecated,
        we have to resort to lower lever commands.
        https://lazka.github.io/pgi-docs/#NM-1.0/classes/Client.html#NM.Client.connectivity_check_set_enabled

        This change is necessary since if this feature is enabled,
        dummy connection are inflated with a value of 20000.

        https://developer-old.gnome.org/NetworkManager/stable/NetworkManager.conf.html
        (see under `connectivity section`)
        """
        if Version(self._nm_client.get_version()) < Version("1.24.0"):
            # NM.Client.connectivity_check_set_enabled is deprecated since version 1.22
            # but the replacement method is only available in version 1.24.
            return self._run_on_glib_loop_thread(
                self._nm_client.connectivity_check_set_enabled, False
            )

        return self._dbus_set_property(
            object_path="/org/freedesktop/NetworkManager",
            interface_name="org.freedesktop.NetworkManager",
            property_name="ConnectivityCheckEnabled",
            value=GLib.Variant("b", False),
            timeout_msec=-1,
            cancellable=None
        )

    def _dbus_set_property(  # pylint: disable=too-many-arguments
            self, *userdata, object_path: str, interface_name: str, property_name: str,
            value: GLib.Variant, timeout_msec: int = -1,
            cancellable: Gio.Cancellable = None,
    ) -> Future:
        """Set NM properties since dedicated methods have been deprecated deprecated.
        Source: https://lazka.github.io/pgi-docs/#NM-1.0/classes/Client.html"""  # noqa

        future = _create_future()

        def _on_property_set(nm_client, res, _user_data):
            if not nm_client or not res or not nm_client.dbus_set_property_finish(res):
                future.set_exception(
                    RuntimeError(
                        f"Error disabling network connectivity check: {nm_client=}, {res=}"
                    )
                )
                return

            future.set_result(None)

        def _set_property_async():
            self._assert_running_on_glib_loop_thread()
            self._nm_client.dbus_set_property(
                object_path, interface_name, property_name,
                value, timeout_msec, cancellable, _on_property_set,
                userdata
            )

        self._run_on_glib_loop_thread(_set_property_async).result()

        return future



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import asyncio
from os.path import basename
from threading import Event
from typing import Optional

from proton.session import Session, FormData, FormField
from proton.session.api import Fido2Assertion, Fido2AssertionParameters

from proton.vpn import logging
from proton.vpn.session.account import VPNAccount
from proton.vpn.session.fetcher import (VPNSessionFetcher, EndpointVersion)
from proton.vpn.session.client_config import ClientConfig
from proton.vpn.session.credentials import VPNSecrets
from proton.vpn.session.dataclasses import LoginResult, BugReportForm, VPNCertificate, VPNLocation
from proton.vpn.session.exceptions import VPNAccountDecodeError, ServerListDecodeError
from proton.vpn.session.servers.logicals import ServerList
from proton.vpn.session.feature_flags_fetcher import FeatureFlags
from proton.vpn.session.u2f_interaction import UserInteraction

logger = logging.getLogger(__name__)

BINARY_SERVER_STATUS = "BinaryServerStatus"


class VPNSession(Session):
    """
    Augmented Session that provides helpers to a persistent offline keyring
    access to user account information available from the PROTON VPN REST API.

    Usage example:

    .. code-block::

        from proton.vpn.session import VPNSession
        from proton.sso import ProtonSSO

        sso = ProtonSSO()
        session=sso.get_session(username, override_class=VPNSession)

        session.authenticate('USERNAME','PASSWORD')

        if session.authenticated:
            pubkey_credentials = session.vpn_account.vpn_credentials.pubkey_credentials
            wireguard_private_key = pubkey_credentials.wg_private_key
            api_pem_certificate = pubkey_credentials.certificate_pem

    """

    BUG_REPORT_ENDPOINT = "/core/v4/reports/bug"

    def __init__(
            self, *args,
            fetcher: Optional[VPNSessionFetcher] = None,
            vpn_account: Optional[VPNAccount] = None,
            server_list: Optional[ServerList] = None,
            client_config: Optional[ClientConfig] = None,
            feature_flags: Optional[FeatureFlags] = None,
            **kwargs
    ):  # pylint: disable=too-many-arguments
        self._fetcher = fetcher or VPNSessionFetcher(session=self)

        try:
            # pylint: disable=import-outside-toplevel
            from proton.vpn.session.u2f import U2FKeys
            self._u2f_keys = U2FKeys()
        except ImportError as error:
            self._u2f_keys = None
            logger.warning(error)

        self._vpn_account = vpn_account
        self._server_list = server_list
        self._client_config = client_config
        self._feature_flags = feature_flags
        super().__init__(*args, **kwargs)

    @property
    def loaded(self) -> bool:
        """:returns: whether the VPN session data was already loaded or not."""
        return self._vpn_account \
            and self._server_list \
            and self._client_config \
            and self._feature_flags

    def __setstate__(self, data):
        """This method is called when deserializing the session."""
        # It might be useful to load feature flags cache/defaults, even in logged out state.
        self._feature_flags = self._fetcher.load_feature_flags_from_cache()

        if 'vpn' in data:
            try:
                self._vpn_account = VPNAccount.from_dict(data['vpn'])
            except VPNAccountDecodeError:
                logger.warning("VPN account could not be deserialized", exc_info=True)

            try:
                self._server_list = self._fetcher.load_server_list_from_cache()
            except ServerListDecodeError:
                logger.warning("Server list could not be deserialized", exc_info=True)

            self._client_config = self._fetcher.load_client_config_from_cache()

        super().__setstate__(data)

    def __getstate__(self):
        """This method is called to retrieve the session data to be serialized in the keyring."""
        state = super().__getstate__()

        if state and self._vpn_account:
            state['vpn'] = self._vpn_account.to_dict()

        # Note the server list is not persisted to the keyring

        return state

    async def login(self, username: str, password: str) -> LoginResult:
        """
        Logs the user in.
        :returns: the login result, indicating whether it was successful
        and whether 2FA is required or not.
        """
        if self.logged_in:
            return LoginResult(success=True, authenticated=True, twofa_required=False)

        if not await self.async_authenticate(username, password):
            return LoginResult(success=False, authenticated=False, twofa_required=False)

        if self.needs_twofa:
            return LoginResult(success=False, authenticated=True, twofa_required=True)

        return LoginResult(success=True, authenticated=True, twofa_required=False)

    async def provide_2fa_code(self, code: str) -> LoginResult:
        """
        Submits the 2FA code.
        :returns: whether the 2FA was successful or not.
        """
        valid_code = await super().async_validate_2fa_code(code)

        if not valid_code:
            return LoginResult(success=False, authenticated=True, twofa_required=True)

        return LoginResult(success=True, authenticated=True, twofa_required=False)

    @property
    def fido2_lib_available(self) -> Fido2AssertionParameters:
        """
        :returns: whether the expected FIDO2 library is available or not.
        """
        return bool(self._u2f_keys)

    async def generate_2fa_fido2_assertion(
            self,
            user_interaction: Optional[UserInteraction] = None,
            cancel_assertion: Optional[Event] = None
    ) -> Fido2Assertion:
        """
        Scans for U2F/FIDO2 keys and generates a FIDO2 assertion.

        :param user_interaction: object handling any required user interaction
            while generating the assertion.
        :param cancel_assertion: optional event that can be set to cancel the
        fido 2 assertion process.
        :returns: the generated FIDO2 assertion.
        """
        if not self.fido2_lib_available:
            raise RuntimeError("U2F/FIDO2 support is not available on this platform")

        return await self._u2f_keys.scan_keys_and_get_assertion(
            self, user_interaction, cancel_assertion
        )

    async def provide_2fa_fido2(self, fido2_assertion: Fido2Assertion) -> LoginResult:
        """
        Submits the 2FA FIDO2 assertion.
        :returns: whether the 2FA was successful or not.
        """
        if not self.fido2_lib_available:
            raise RuntimeError("U2F/FIDO2 support is not available on this platform")

        valid_assertion = await super().async_validate_2fa_fido2(fido2_assertion)

        if not valid_assertion:
            return LoginResult(success=False, authenticated=True, twofa_required=True)

        return LoginResult(success=True, authenticated=True, twofa_required=False)

    async def logout(self, no_condition_check=False, additional_headers=None) -> bool:
        """
        Log out and reset session data.
        """
        result = await super().async_logout(no_condition_check, additional_headers)
        self._vpn_account = None
        self._server_list = None
        self._client_config = None
        self._feature_flags = None
        self._fetcher.clear_cache()
        return result

    @property
    def logged_in(self) -> bool:
        """
        :returns: whether the user already logged in or not.
        """
        return self.authenticated and not self.needs_twofa

    async def fetch_session_data(self, features: Optional[dict] = None):
        """
        Fetches the required session data from Proton's REST APIs.
        """

        # We have to use `no_condition_check=True` with `_requests_lock`
        # because otherwise all requests after that will be blocked
        # until the lock created by `_requests_lock` is released.
        # Since the previous lock is only released at the end of the try/except/finally the
        # requests will never be executed, thus blocking and never releasing the lock.

        # Each request in `proton.session.api.Session` already creates and holds the lock by itself,
        # but the problem here is that we want to add additional data to be stored to the keyring.
        # Thus we need to resort to some manual
        # triggering of `_requests_lock` and `_requests_unlock`.
        # The former caches keyring data to memory while the latter does three different things:
        #   1. It checks if the new data is different from the old one
        #   2. If they are different then it proceeds to delete old one from keyring
        #   3. Add new data to the keyring
        # So if we want to add additional data to the keyring, as in VPN relevant data,
        # we must ensure that we always call `_requests_unlock()` after any requests
        # because this is currently the only way to store data that is attached
        # to a specific account.

        # So the consequence for passing `no_condition_check=True` is that the keyring data will
        # not get cached to memory, for later to be compared (as previously described).
        # This means that later when the comparison will be made, the "old" data will just be empty,
        # forcing it to always be replaced by the new data to keyring. Thus this solution is just a
        # temporary hack until a better approach is found.

        # For further clarification on how these methods see the following, in the specified order:
        #     `proton.session.api.Session._requests_lock`
        #     `proton.sso.sso.ProtonSSO._acquire_session_lock`
        #     `proton.session.api.Session._requests_unlock`
        #     `proton.sso.sso.ProtonSSO._release_session_lock`

        self._requests_lock(no_condition_check=True)
        try:
            secrets = (
                VPNSecrets(
                    ed25519_privatekey=self._vpn_account.vpn_credentials
                    .pubkey_credentials.ed_255519_private_key
                )
                if self._vpn_account
                else VPNSecrets()
            )

            vpninfo, certificate, location, client_config = await asyncio.gather(
                self._fetcher.fetch_vpn_info(),
                self._fetcher.fetch_certificate(
                    client_public_key=secrets.ed25519_pk_pem, features=features),
                self._fetcher.fetch_location(),
                self._fetcher.fetch_client_config(),
            )

            self._vpn_account = VPNAccount(
                vpninfo=vpninfo, certificate=certificate, secrets=secrets, location=location
            )
            self._client_config = client_config

            # The feature flags must be fetched before the server list,
            # since the server list can be fetched differently depending on
            # what feature flags are enabled.
            self._feature_flags = await self._fetcher.fetch_feature_flags()

            # The server list should be retrieved after the VPNAccount object
            # has been created, since it requires the location, and it should
            # be retrieved after the feature flags have been fetched, since it
            # depends in them for chosing the fetch method.
            await self.fetch_server_list()

        finally:
            # IMPORTANT: apart from releasing the lock, _requests_unlock triggers the
            # serialization of the session to the keyring.
            self._requests_unlock()

    async def fetch_certificate(self, features: Optional[dict] = None) -> VPNCertificate:
        """Fetches new certificate from API."""

        self._requests_lock(no_condition_check=True)
        try:
            secrets = (
                VPNSecrets(
                    ed25519_privatekey=self._vpn_account.vpn_credentials
                    .pubkey_credentials.ed_255519_private_key
                )
            )
            new_certificate = await self._fetcher.fetch_certificate(
                client_public_key=secrets.ed25519_pk_pem,
                features=features
            )
            self._vpn_account.set_certificate(new_certificate)

            return new_certificate
        finally:
            self._requests_unlock()

    @property
    def vpn_account(self) -> VPNAccount:
        """
        Information related to the VPN user account.
        If it was not loaded yet then None is returned instead.
        """
        return self._vpn_account

    def set_location(self, location: VPNLocation):
        """Set new location data and store it."""
        self._requests_lock(no_condition_check=False)
        try:
            self._vpn_account.location = location
        finally:
            self._requests_unlock()

    def _serverlist_endpoint_version(self) -> EndpointVersion:
        """Returns the endpoint version to be used for server list fetching."""
        if self._feature_flags.get(BINARY_SERVER_STATUS):
            return EndpointVersion.V2
        return EndpointVersion.V1

    async def fetch_server_list(self) -> ServerList:
        """
        Fetches the server list from the REST API.
        """
        self._server_list = await self._fetcher.fetch_server_list(
            self._serverlist_endpoint_version()
        )
        return self._server_list

    @property
    def server_list(self) -> ServerList:
        """The current server list."""
        return self._server_list

    async def update_server_loads(self) -> ServerList:
        """
        Fetches the server loads from the REST API and updates the current
        server list with them.
        """
        self._server_list = await self._fetcher.update_server_loads(
            self._serverlist_endpoint_version()
        )
        return self._server_list

    async def fetch_client_config(self) -> ClientConfig:
        """Fetches the client configuration from the REST api."""
        self._client_config = await self._fetcher.fetch_client_config()
        return self._client_config

    @property
    def client_config(self) -> ClientConfig:
        """The current client configuration."""
        return self._client_config

    async def fetch_feature_flags(self) -> FeatureFlags:
        """Fetches API features that dictates which features are to be enabled or not."""
        self._feature_flags = await self._fetcher.fetch_feature_flags()
        return self._feature_flags

    @property
    def feature_flags(self) -> FeatureFlags:
        """Fetches general client configuration to connect to VPN servers."""
        if self._feature_flags is None:
            return FeatureFlags.default()
        return self._feature_flags

    async def submit_bug_report(self, bug_report: BugReportForm):
        """Submits a bug report to customer support."""
        data = FormData()
        data.add(FormField(name="OS", value=bug_report.os))
        data.add(FormField(name="OSVersion", value=bug_report.os_version))
        data.add(FormField(name="Client", value=bug_report.client))
        data.add(FormField(name="ClientVersion", value=bug_report.client_version))
        data.add(FormField(name="ClientType", value=bug_report.client_type))
        data.add(FormField(name="Title", value=bug_report.title))
        data.add(FormField(name="Description", value=bug_report.description))
        data.add(FormField(name="Username", value=bug_report.username))
        data.add(FormField(name="Email", value=bug_report.email))

        if self._vpn_account:
            location = self._vpn_account.location
            data.add(FormField(name="ISP", value=location.ISP))
            data.add(FormField(name="Country", value=location.Country))

        for i, attachment in enumerate(bug_report.attachments):
            data.add(FormField(
                name=f"Attachment-{i}", value=attachment,
                filename=basename(attachment.name)
            ))

        return await self.async_api_request(
            endpoint=VPNSession.BUG_REPORT_ENDPOINT, data=data
        )



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations

from typing import Sequence, TYPE_CHECKING

from proton.vpn.session.credentials import VPNPubkeyCredentials, VPNSecrets
from proton.vpn.session.dataclasses import (
    VPNSettings, VPNLocation, VPNCertificate,
    VPNCredentials, VPNUserPassCredentials
)
from proton.vpn.session.exceptions import VPNAccountDecodeError

if TYPE_CHECKING:
    from proton.vpn.session.dataclasses import APIVPNSession


class VPNAccount:
    """
    This class is responsible to encapsulate all user vpn account information,
    including credentials (private keys, vpn user and password).
    """

    def __init__(
        self, vpninfo: VPNSettings, certificate: VPNCertificate,
        secrets: VPNSecrets, location: VPNLocation
    ):
        self._vpninfo = vpninfo
        self._certificate = certificate
        self._secrets = secrets
        self.location = location

    @staticmethod
    def from_dict(dict_data: dict) -> VPNAccount:
        """Creates a VPNAccount instance from the specified
            dictionary for deserialization purposes."""
        try:
            return VPNAccount(
                vpninfo=VPNSettings.from_dict(dict_data['vpninfo']),
                certificate=VPNCertificate.from_dict(dict_data['certificate']),
                secrets=VPNSecrets.from_dict(dict_data['secrets']),
                location=VPNLocation.from_dict(dict_data['location'])
            )
        except Exception as exc:
            raise VPNAccountDecodeError("Invalid VPN account") from exc

    def set_certificate(self, new_certificate: VPNCertificate):
        """Set new certificate.
        This affects only when asking for `vpn_credentials` property
        as it's built on the fly.
        """
        self._certificate = new_certificate

    def to_dict(self) -> dict:
        """
        Returns this object as a dictionary for serialization purposes.
        """
        return {
            "vpninfo": self._vpninfo.to_dict(),
            "certificate": self._certificate.to_dict(),
            "secrets": self._secrets.to_dict(),
            "location": self.location.to_dict()
        }

    @property
    def plan_name(self) -> str:
        """
        :return: str `PlanName` value of the account from :class:`api_data.VPNInfo` in
            Non-human readable format.
        """
        return self._vpninfo.VPN.PlanName

    @property
    def plan_title(self) -> str:
        """
        :return: str `PlanName` value of the account from :class:`api_data.VPNInfo`,
            Human readable format, thus if you intend to display the plan
            to the user use this one instead of :class:`VPNAccount.plan_name`.
        """
        return self._vpninfo.VPN.PlanTitle

    @property
    def max_tier(self) -> int:
        """
        :return: int `Maxtier` value of the account from :class:`api_data.VPNInfo`.
        """
        return self._vpninfo.VPN.MaxTier

    @property
    def max_connections(self) -> int:
        """
        :return: int the `MaxConnect` value of the account from :class:`api_data.VPNInfo`.
        """
        return self._vpninfo.VPN.MaxConnect

    @property
    def delinquent(self) -> bool:
        """
        :return: bool if the account is delinquent,
            based the value from :class:`api_data.VPNSettings`.
        """
        return self._vpninfo.Delinquent > 2

    @property
    def active_connections(self) -> Sequence["APIVPNSession"]:
        """
        :return: the list of active VPN session of the authenticated user on the infra.
        """
        raise NotImplementedError

    @property
    def vpn_credentials(self) -> VPNCredentials:
        """ Return :class:`protonvpn.vpnconnection.interfaces.VPNCredentials` to
            provide an interface readily usable to
            instantiate a :class:`protonvpn.vpnconnection.VPNConnection`.
        """
        return VPNCredentials(
            userpass_credentials=VPNUserPassCredentials(
                username=self._vpninfo.VPN.Name,
                password=self._vpninfo.VPN.Password
            ),
            pubkey_credentials=VPNPubkeyCredentials(
                api_certificate=self._certificate,
                secrets=self._secrets,
                strict=True
            )
        )



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from datetime import timedelta

from proton.vpn.core.refresher.scheduler import RunAgain
from proton.vpn.core.session_holder import SessionHolder
from proton.vpn.session.client_config import ClientConfig

from proton.vpn import logging
from proton.session.exceptions import (
    ProtonAPINotReachable, ProtonAPINotAvailable,
    ProtonAPIError
)

logger = logging.getLogger(__name__)

# pylint: disable=R0801


class ClientConfigRefresher:
    """
    Service in charge of refreshing VPN client configuration data.
    """
    def __init__(self, session_holder: SessionHolder):
        super().__init__()
        self._session_holder = session_holder

    @property
    def _session(self):
        return self._session_holder.session

    @property
    def initial_refresh_delay(self):
        """Returns the initial delay before the first refresh."""
        return self._session.client_config.seconds_until_expiration

    async def update_if_necessary(self):
        """Fetches the new client config if the current one has expired"""
        if not self._session.client_config.is_expired:
            return  # no need to update too early

        try:
            await self._session.fetch_client_config()
        except ProtonAPIError as error:
            if error.http_code != 429:
                raise

            logger.warning(f"Client config refresh failed: {error}")
        except (ProtonAPINotReachable, ProtonAPINotAvailable) as error:
            logger.warning(f"Client config refresh failed: {error}")
        except Exception:
            logger.error(
                "Client config refresh failed unexpectedly. "
                "Stopping client config refresh."
            )
            raise

    async def refresh(self) -> RunAgain:
        """Fetches the new client configuration from the REST API."""
        try:
            new_client_config = await self._session.fetch_client_config()
            next_refresh_delay = new_client_config.seconds_until_expiration
        except ProtonAPIError as error:
            if error.http_code != 429:
                raise

            logger.warning(f"Client config refresh failed: {error}")
            next_refresh_delay = ClientConfig.get_refresh_interval_in_seconds()
        except (ProtonAPINotReachable, ProtonAPINotAvailable) as error:
            logger.warning(f"Client config refresh failed: {error}")
            next_refresh_delay = ClientConfig.get_refresh_interval_in_seconds()
        except Exception:
            logger.error(  # nosec B311 # noqa: E501 # pylint: disable=line-too-long # nosemgrep: python.lang.best-practice.logging-error-without-handling.logging-error-without-handling
                "Client config refresh failed unexpectedly. "
                "Stopping client config refresh."
            )
            raise

        logger.info(
            f"Next client config refresh scheduled in "
            f"{timedelta(seconds=next_refresh_delay)}"
        )

        return RunAgain.after_seconds(next_refresh_delay)



"""
This module manages the Proton VPN general settings.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""

from __future__ import annotations

from dataclasses import dataclass
import os

from proton.vpn import logging
from proton.utils.environment import VPNExecutionEnvironment
from proton.vpn.core.cache_handler import CacheHandler
from proton.vpn.killswitch.interface import KillSwitchState
from proton.vpn.core.settings.custom_dns import CustomDNS
from proton.vpn.core.settings.features import Features


logger = logging.getLogger(__name__)


SETTINGS = os.path.join(
    VPNExecutionEnvironment().path_config,
    "settings.json"
)


DEFAULT_PROTOCOL = "wireguard"
DEFAULT_KILLSWITCH = KillSwitchState.OFF.value
DEFAULT_ANONYMOUS_CRASH_REPORTS = True


@dataclass
class Settings:
    """Contains general settings."""
    protocol: str
    killswitch: int
    custom_dns: CustomDNS
    ipv6: bool
    anonymous_crash_reports: bool
    features: Features

    @staticmethod
    def from_dict(data: dict, user_tier: int) -> Settings:
        """Creates and returns `Settings` from the provided dict."""
        default = Settings.default(user_tier)

        features = data.get("features")
        features = Features.from_dict(
            features, user_tier) if features else default.features
        custom_dns = data.get("custom_dns")
        custom_dns = CustomDNS.from_dict(
            custom_dns) if custom_dns else default.custom_dns

        return Settings(
            protocol=data.get("protocol", default.protocol),
            killswitch=data.get("killswitch", default.killswitch),
            custom_dns=custom_dns,
            ipv6=data.get("ipv6", default.ipv6),
            anonymous_crash_reports=data.get(
                "anonymous_crash_reports",
                default.anonymous_crash_reports
            ),
            features=features
        )

    def to_dict(self) -> dict:
        """Converts the class to dict."""
        return {
            "protocol": self.protocol,
            "killswitch": self.killswitch,
            "custom_dns": self.custom_dns.to_dict(),
            "ipv6": self.ipv6,
            "anonymous_crash_reports": self.anonymous_crash_reports,
            "features": self.features.to_dict(),
        }

    @staticmethod
    def default(user_tier: int) -> Settings:
        """Creates and returns `Settings` from default configurations."""
        return Settings(
            protocol=DEFAULT_PROTOCOL,
            killswitch=DEFAULT_KILLSWITCH,
            custom_dns=CustomDNS.default(),
            ipv6=True,
            anonymous_crash_reports=DEFAULT_ANONYMOUS_CRASH_REPORTS,
            features=Features.default(user_tier),
        )


class SettingsPersistence:
    """Persists user settings"""

    def __init__(self, cache_handler: CacheHandler = None):
        self._cache_handler = cache_handler or CacheHandler(SETTINGS)
        self._settings = None
        self._settings_are_default = True

    def get(self, user_tier: int) -> Settings:
        """Load the user settings, either the ones stored on disk or getting
        default based on tier"""

        if self._settings is not None:
            return self._settings

        raw_settings = self._cache_handler.load()
        if raw_settings is None:
            self._settings = Settings.default(user_tier)
        else:
            self._settings = Settings.from_dict(raw_settings, user_tier)
            self._settings_are_default = False

        return self._settings

    def save(self, settings: Settings):
        """Store settings to disk."""
        self._cache_handler.save(settings.to_dict())
        self._settings = settings
        self._settings_are_default = False

    def delete(self):
        """Deletes the file stored on disk containing the settings
        and resets internal settings property."""
        self._cache_handler.remove()
        self._settings = None
        self._settings_are_default = True



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from dataclasses import dataclass
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from proton.vpn.session.credentials import VPNPubkeyCredentials

# pylint: disable=invalid-name


@dataclass
class VPNUserPassCredentials:
    """ Class responsible to hold vpn user/password credentials for authentication
    """
    username: str
    password: str


@dataclass
class VPNCredentials:
    """ Interface to :class:`proton.vpn.connection.interfaces.VPNCredentials`
        See :attr:`proton.vpn.session.VPNSession.vpn_account.vpn_credentials` to get one.
    """
    userpass_credentials: VPNUserPassCredentials
    pubkey_credentials: VPNPubkeyCredentials



"""
Exceptions raised by the VPN connection module.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""


class VPNConnectionError(Exception):
    """Base class for VPN specific exceptions"""
    def __init__(self, message, additional_context=None):
        self.message = message
        self.additional_context = additional_context
        super().__init__(self.message)


class AuthenticationError(VPNConnectionError):
    """When server answers with auth_denied this exception is thrown.

    In many cases, an auth_denied can be thrown for multiple reasons, thus it's up to
    the user to decide how to proceed further.
    """


class ConnectionTimeoutError(VPNConnectionError):
    """When a connection takes too long to connect, this exception will be thrown."""


class MissingBackendDetails(VPNConnectionError):
    """When no VPN backend is found (NetworkManager, Native, etc) then this exception is thrown.

    In rare cases where it can happen that a user has some default packages installed, where the
    services for those packages are actually not running. Ie:
    NetworkManager is installed but not running and for some reason we can't access native backend,
    thus this exception is thrown as we can't do anything.
    """


class MissingProtocolDetails(VPNConnectionError):
    """
    When no VPN protocol is found (OpenVPN, Wireguard, IKEv2, etc) then this exception is thrown.
    """


class ConcurrentConnectionsError(VPNConnectionError):
    """
    Multiple concurrent connections were found, even though only one is allowed at a time.
    """


class FeatureError(VPNConnectionError):
    """
    Feature errors are thrown when the server fails to set the requested connection feature.
    """


class FeaturePolicyError(FeatureError):
    """
    Policy errors happen when the server fails to set the requested connection feature,
    either because the user doesn't have the rights to do so or because of
    server-side issues.
    """


class FeatureSyntaxError(FeatureError):
    """
    Syntax errors are programming errors, meaning that what we the request to set the
    connection feature is incorrect, ie: passing wrong/non-existent values, format is
    incorrect, etc.
    """


class HardJailedTwoFAError(VPNConnectionError):
    """
    When a user is hard jailed due to 2FA issues, this exception is thrown.
    """



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from typing import List
from dataclasses import dataclass
from proton.vpn.session.utils import Serializable

# pylint: disable=invalid-name


@dataclass
class APIVPNSession(Serializable):  # pylint: disable=missing-class-docstring
    SessionID: str
    ExitIP: str
    Protocol: str

    @staticmethod
    def _deserialize(dict_data: dict) -> APIVPNSession:
        return APIVPNSession(
            SessionID=dict_data["SessionID"],
            ExitIP=dict_data["ExitIP"],
            Protocol=dict_data["Protocol"]
        )


@dataclass
class VPNSessions(Serializable):
    """ The list of active VPN session of an account on the infra """
    Sessions: List[APIVPNSession]

    def __len__(self):
        return len(self.Sessions)

    @staticmethod
    def _deserialize(dict_data: dict) -> VPNSessions:
        session_list = [APIVPNSession.from_dict(value) for value in dict_data['Sessions']]
        return VPNSessions(Sessions=session_list)



"""
Copyright (c) 2024 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import asyncio
from collections.abc import Callable
from proton.vpn.local_agent import (  # pylint: disable=no-name-in-module, import-error
    Listener, AgentFeatures, Status
)
from proton.vpn.session.credentials import VPNPubkeyCredentials

from proton.vpn import logging

logger = logging.getLogger(__name__)


class AgentListener:
    """Listens for local agent messages asynchronously."""

    def __init__(self, on_status: Callable[[Status], None], on_error: Callable[[Exception], None]):
        self._listener = None
        self._future = None
        self._on_status_callback = on_status
        self._on_error_callback = on_error

    async def connect(self, domain: str, credentials: VPNPubkeyCredentials):
        """Establishes a connection to local agent server.

        :param domain: the domain of the VPN server.
        :param credentials: object that contains all necessary user credentials.
        :raises ExpiredCertificateError: when the certificate provided is expired or invalid.
        :raises TimeoutError: when the connection to LA server times out.
        :raises Exception: any other exceptions.
        """
        private_key = credentials.get_ed25519_sk_pem()
        certificate = credentials.certificate_pem
        self._listener = await Listener.connect(domain, private_key,
                                                certificate)

    async def listen(self):
        """Starts the background process of listening to incoming messages from LA."""
        self._future = self._listener.listen(
            self._on_status_callback, self._on_error_callback
        )
        await self._wait_for_it(self._future)

    async def request_features(self, features: AgentFeatures):
        """Requests the features to be set on the current VPN connection."""
        await self._listener.request_features(features)

    async def stop(self):
        """Stops reading incoming messages from LA."""
        if not self._future or self._future.done():
            return

        self._future.cancel()
        await self._wait_for_it(self._future)
        logger.info("Agent listener successfully stopped.")

    @property
    def is_running(self) -> bool:
        """Returns if the future is still running."""
        return self._future and not self._future.done()

    @staticmethod
    async def _wait_for_it(future):
        try:
            await future
        except asyncio.CancelledError:
            pass



"""
List of exceptions raised in this package.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Optional, Union

from proton.loader import Loader

from proton.vpn import logging
from proton.vpn.core.settings import SplitTunnelingConfig

logger = logging.getLogger(__name__)


class SplitTunneling(ABC):
    """Defines the interface to create a new split tunneling client.
    """
    def __init__(self, uid: int):
        self._uid: int = uid

    @staticmethod
    async def get(uid: int) -> Optional[SplitTunneling]:
        """
        Returns the split tunneling implementation.

        :param uid: unix user ID
        """
        try:
            split_tunneling_class = Loader.get(
                type_name="split_tunneling"
            )
        except RuntimeError:
            logger.warning("Split tunneling backend not found")
            return None

        return await split_tunneling_class.build(uid)

    @staticmethod
    @abstractmethod
    async def build(uid: int) -> SplitTunneling:
        """Builds and returns the split tunneling instance."""

    @abstractmethod
    async def set_config(self, config: SplitTunnelingConfig) -> None:
        """Sets a new config for instance uid.

        Args:
            config (SplitTunnelingConfig): the object containing the data
            uid (int): uid that the config has to be stored for
        """

    @abstractmethod
    async def get_config(self) -> Optional[SplitTunnelingConfig]:
        """Returns config for instance uid.

        Args:
            uid (int): uid to get the data for

        Returns:
            SplitTunnelingConfig: contains data stored for the specified uid
        """

    @abstractmethod
    async def clear_config(self) -> None:
        """Clears config stored for instance uid.

        Args:
            uid (int): uid that data is to be cleared for
        """

    @abstractmethod
    async def get_all_configs(
            self
    ) -> Union[
        list[tuple[int, SplitTunnelingConfig]],
        list
    ]:
        """Clears data stored for the specified uid.
        """

    @classmethod
    @abstractmethod
    def _get_priority(cls) -> int:
        """
        Priority of the split tunneling implementation.

        To be implemented by subclasses.
        """

    @classmethod
    @abstractmethod
    def _validate(cls) -> bool:
        """
        Determines whether the split tunneling connection implementation is valid or not.
        To be implemented by subclasses.
        """



"""VPN connection enums.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""

from enum import auto, Enum, IntEnum


class ConnectionStateEnum(IntEnum):
    """VPN connection states."""
    DISCONNECTED = 0
    CONNECTING = 1
    CONNECTED = 2
    DISCONNECTING = 3
    ERROR = 4


class StateMachineEventEnum(Enum):
    """VPN connection events."""
    INITIALIZED = auto()
    UP = auto()
    DOWN = auto()
    CONNECTED = auto()
    DISCONNECTED = auto()
    TIMEOUT = auto()
    AUTH_DENIED = auto()
    TUNNEL_SETUP_FAILED = auto()
    RETRY = auto()
    UNEXPECTED_ERROR = auto()
    DEVICE_DISCONNECTED = auto()
    CERTIFICATE_EXPIRED = auto()
    MAXIMUM_SESSIONS_REACHED = auto()
    UNHANDLED_ERROR = auto()
    TWOFA_REQUIRED = auto()


class KillSwitchSetting(IntEnum):
    """Kill switch setting values."""
    OFF = 0
    ON = 1
    PERMANENT = 2



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from typing import List
from dataclasses import dataclass


@dataclass
class ProtocolPorts:
    """Dataclass for ports.
    These ports are mainly used for establishing VPN connections.
    """
    udp: List
    tcp: List
    tls: List

    @staticmethod
    def from_dict(ports: dict) -> ProtocolPorts:
        """Creates ProtocolPorts object from data."""
        # The lists are copied to avoid side effects if the dict is modified.
        return ProtocolPorts(
            udp=ports["UDP"].copy(),
            tcp=ports["TCP"].copy(),
            tls=ports.get("TLS", []).copy()
        )



"""
Base class for VPN connections created with NetworkManager.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import asyncio
import logging
from typing import Optional

from proton.loader import Loader
from proton.vpn.connection import VPNConnection, events, states
from proton.vpn.connection.events import EventContext, Event
from proton.vpn.connection.vpnconfiguration import VPNConfiguration
from proton.vpn.connection.states import StateContext


from proton.vpn.backend.networkmanager.core.nmclient import (NM, NMClient, GLib)
from proton.vpn.backend.networkmanager.core import tcpcheck

logger = logging.getLogger(__name__)


class LinuxNetworkManager(VPNConnection):
    """VPNConnections based on Linux Network Manager backend.

    The LinuxNetworkManager connection backend could return VPNConnection
    instances based on protocols such as OpenVPN, IKEv2 or Wireguard.

    To do this, this class needs to be extended so that the subclass
    initializes the NetworkManager connection appropriately.
    """
    SIGNAL_NAME = "vpn-state-changed"
    backend = "linuxnetworkmanager"

    def __init__(self, *args, nm_client: NMClient = None, **kwargs):
        self.__nm_client = nm_client
        self._asyncio_loop = asyncio.get_running_loop()
        self._cancelled = False
        super().__init__(*args, **kwargs)

    @property
    def nm_client(self):
        """Returns the NetworkManager client."""
        if not self.__nm_client:
            self.__nm_client = NMClient()

        return self.__nm_client

    @classmethod
    def factory(cls, protocol: str = None):
        """Returns the VPN connection implementation class
         for the specified protocol."""
        return Loader.get(LinuxNetworkManager.backend, class_name=protocol)

    async def start(self):
        """
        Starts a VPN connection using NetworkManager.
        """
        # The VPN connection is started only if at least one of the TCP ports of the server to
        # connect to is open. The reason for doing this check is that, after introducing the
        # dummy kill switch network interface, the VPN connection backend tries to use it
        # to establish the VPN connection.
        self._cancelled = False

        server_reachable = await tcpcheck.is_any_port_reachable(
            self._vpnserver.server_ip,
            self._vpnserver.openvpn_ports.tcp
        )

        if not server_reachable:
            logger.info("VPN server NOT reachable.")
            self._notify_subscribers(events.Timeout(EventContext(connection=self)))
            return

        logger.info("VPN server REACHABLE.")

        if self._cancelled:
            logger.info("Connection cancelled.")
            self._notify_subscribers(events.Disconnected(EventContext(connection=self)))
            return

        future_connection = self.setup()  # Creates the network manager connection.
        loop = asyncio.get_running_loop()
        try:
            connection = await loop.run_in_executor(None, future_connection.result)
        except GLib.GError:
            logger.exception("Error adding NetworkManager connection.")
            self._notify_subscribers(
                events.TunnelSetupFailed(
                    context=EventContext(
                        connection=self,
                        error=NM.VpnConnectionStateReason.NONE.real
                    )
                )
            )
            return

        if self._cancelled:
            logger.info("Connection cancelled.")
            await self.remove_connection(connection)
            self._notify_subscribers(events.Disconnected(EventContext(connection=self)))
            return

        try:
            future_vpn_connection = self.nm_client.start_connection_async(connection)
            vpn_connection = await loop.run_in_executor(
                None, future_vpn_connection.result
            )
            # Start listening for vpn state changes.
            vpn_connection.connect(
                self.SIGNAL_NAME,
                self._on_state_changed
            )
        except GLib.GError:
            logger.exception("Error starting NetworkManager connection.")
            self._notify_subscribers(
                events.TunnelSetupFailed(
                    context=EventContext(
                        connection=self,
                        error=NM.VpnConnectionStateReason.NONE.real
                    )
                )
            )
            await self.remove_connection(connection)

    async def stop(self, connection=None):
        """Stops the VPN connection."""
        # We directly remove the connection to avoid leaking NM connections.
        if not self._is_nm_connection_active():
            self._notify_subscribers(
                events.Disconnected(EventContext(connection=self))
            )

        connection = connection or self._get_nm_connection()
        if not connection:
            # It can happen that a connection is stopped while checking if the server
            # is reachable, but before the underlying NM connection is created.
            # In that case we flag it as cancelled, so the creation of the underlying
            # NM connection is skipped.
            self._cancelled = True
        else:
            await self.remove_connection(connection)

    async def remove_connection(self, connection=None):
        """Removes the VPN connection."""
        connection = connection or self._get_nm_connection()
        if not connection:
            return

        future = self.nm_client.remove_connection_async(connection)
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(None, future.result)

        self._unique_id = None

    async def remove_persistence(self):
        await super().remove_persistence()
        await self.remove_connection()

    def _notify_subscribers_threadsafe(self, event: Event):
        """
        When notifying subscribers from different thread than then one running the
        asyncio loop, this method must be used for thread-safety reasons.
        """
        self._asyncio_loop.call_soon_threadsafe(
            self._notify_subscribers, event
        )

    def _initialize_persisted_connection(
            self, connection_id: str
    ) -> states.State:
        """Abstract method implementation."""
        context = StateContext(
            event=events.Initialized(EventContext(connection=self)),
            connection=self
        )
        active_connection = self.nm_client.get_active_connection(
            connection_id
        )
        if active_connection:
            active_connection.connect(
                self.SIGNAL_NAME,
                self._on_state_changed
            )
            return states.Connected(context)

        inactive_connection = self.nm_client.get_connection(connection_id)
        if inactive_connection:
            # If the connection is inactive then it means that it unexpectedly dropped.
            # Note that when the user willingly disconnects then the VPN connection is discarded.
            return states.Error(context)

        return states.Disconnected(context)

    def _on_state_changed(
            self, vpn_connection: NM.VpnConnection, state: int, reason: int
    ):
        """
        Use to respond to state changes in the VPN connection, must be
        implemented by the derived class.
        """
        raise NotImplementedError

    def _get_servername(self) -> "str":
        server_name = self._vpnserver.server_name or "Connection"
        return f"ProtonVPN {server_name}"

    def setup(self):
        """
        Every protocol derived from this class has to override this method
        in order to have it working.
        """
        raise NotImplementedError

    @property
    def enable_ipv6_support(self) -> bool:
        """Returns if IPv6 support is enabled or not."""
        return self._vpnserver.has_ipv6_support and self._settings.ipv6

    @classmethod
    def _get_priority(cls):
        return 100

    @classmethod
    def _validate(cls):
        # FIX ME: This should do a validation to ensure that NM can be used
        return True

    @classmethod
    def _plugin_exists(cls, plugin_name: str) -> bool:
        """Checks if there is a NetworkManager plugin with the given name."""
        for i in NM.VpnPluginInfo.list_load():
            if i.get_name() == plugin_name:
                return True
        return False

    def _import_vpn_config(
            self,
            plugin_name: str,
            vpnconfig: VPNConfiguration,
    ) -> NM.SimpleConnection:
        """
            Imports the vpn connection configurations into NM
            and stores the connection on non-volatile memory.

            :return: imported vpn connection
            :rtype: NM.SimpleConnection
        """
        vpn_plugin_list = NM.VpnPluginInfo.list_load()

        connection = None
        with vpnconfig as filename:
            for plugin in vpn_plugin_list:
                if plugin.get_name() == plugin_name:
                    plugin_editor = plugin.load_editor_plugin()
                    # return a NM.SimpleConnection (NM.Connection)
                    # https://lazka.github.io/pgi-docs/NM-1.0/classes/SimpleConnection.html
                    try:
                        connection = plugin_editor.import_(filename)
                        break
                    except GLib.Error:
                        continue

        if connection is None:
            raise NotImplementedError(
                "Support for given configuration is not implemented"
            )

        # https://lazka.github.io/pgi-docs/NM-1.0/classes/Connection.html#NM.Connection.normalize
        connection.normalize()

        return connection

    def _get_nm_connection(self) -> Optional[NM.RemoteConnection]:
        """Get ProtonVPN connection, if there is one."""
        if not self._unique_id:
            return None
        return self.nm_client.get_connection(self._unique_id)

    def _is_nm_connection_active(self):
        if not self._unique_id:
            return False

        return bool(self.nm_client.get_active_connection(self._unique_id))



"""
Interfaces required to be able to establish a VPN connection.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from typing import List, Optional, Protocol
from dataclasses import dataclass


@dataclass
class ProtocolPorts:  # pylint: disable=R0801
    """Dataclass for ports.
    These ports are mainly used for establishing VPN connections.
    """
    udp: List
    tcp: List
    tls: List

    @staticmethod
    def from_dict(ports: dict) -> ProtocolPorts:
        """Creates ProtocolPorts object from data."""
        # The lists are copied to avoid side effects if the dict is modified.
        return ProtocolPorts(
            udp=ports["udp"].copy(),
            tcp=ports["tcp"].copy(),
            tls=ports.get("tls", []).copy()
        )

    def to_dict(self) -> dict:
        """
        Returns a dictionary representation of the object.
        """
        return {
            "udp": self.udp.copy(),
            "tcp": self.tcp.copy(),
            "tls": self.tls.copy()
        }


@dataclass
class VPNServer:  # pylint: disable=too-few-public-methods,too-many-instance-attributes
    """
    Contains the necessary data about the server to connect to.

    Some properties like server_id and server_name are not used to establish
    the connection, but they are required for bookkeeping.
    When the connection is retrieved from persistence, then VPN clients
    can use this information to be able to identify the server that
    the VPN connection was established to. The server name is there mainly
    for debugging purposes.

    Attributes:
        server_ip: server ip to connect to.
        domain: domain to be used for x509 verification.
        x25519pk: x25519 public key for wireguard peer verification.
        wireguard_ports: Dict of WireGuard ports, if the protocol requires them.
        openvpn_ports: Dict of OpenVPN ports, if the protocol requires them.
        server_id: ID of the server to connect to.
        server_name: Name of the server to connect to.
    """
    server_ip: str
    openvpn_ports: ProtocolPorts
    wireguard_ports: ProtocolPorts
    domain: str
    x25519pk: str
    server_id: str
    server_name: str
    has_ipv6_support: bool
    label: str = None

    def __str__(self):
        return f"Server: {self.server_name} / Domain: {self.domain} / " \
               f"IP: {self.server_ip} / OpenVPN Ports: {self.openvpn_ports} / " \
               f"WireGuard Ports: {self.wireguard_ports}"

    @staticmethod
    def from_dict(data: dict) -> VPNServer:
        """
        Creates a VPNServer object from a dictionary.
        """
        return VPNServer(
            server_ip=data["server_ip"],
            openvpn_ports=ProtocolPorts.from_dict(data["openvpn_ports"]),
            wireguard_ports=ProtocolPorts.from_dict(data["wireguard_ports"]),
            domain=data["domain"],
            x25519pk=data["x25519pk"],
            server_id=data["server_id"],
            server_name=data["server_name"],
            has_ipv6_support=data["has_ipv6_support"],
            label=data.get("label")
        )

    def to_dict(self) -> dict:
        """
        Returns a dictionary representation of the object.
        """
        return {
            "server_ip": self.server_ip,
            "openvpn_ports": self.openvpn_ports.to_dict(),
            "wireguard_ports": self.wireguard_ports.to_dict(),
            "domain": self.domain,
            "x25519pk": self.x25519pk,
            "server_id": self.server_id,
            "server_name": self.server_name,
            "has_ipv6_support": self.has_ipv6_support,
            "label": self.label
        }


class VPNPubkeyCredentials(Protocol):  # pylint: disable=too-few-public-methods
    """
    Object that gets certificates and privates keys
    for certificate based connections.

    An instance of this class is to be passed to VPNCredentials.

    Attributes:
        certificate_pem: X509 client certificate in PEM format.
        wg_private_key: wireguard private key in base64 format.
        openvpn_private_key: OpenVPN private key in PEM format.
    """
    certificate_pem: str
    wg_private_key: str
    openvpn_private_key: str


class VPNUserPassCredentials(Protocol):  # pylint: disable=too-few-public-methods
    """Provides username and password for username/password VPN authentication."""
    username: str
    password: str


class VPNCredentials(Protocol):  # pylint: disable=too-few-public-methods
    """
    Credentials are needed to establish a VPN connection.
    Depending on how these credentials are used, one method or the other may be
    irrelevant.

    Limitation:
    You could define only userpass_credentials, though at the cost that you
    won't be able to connect to wireguard (since it's based on certificates)
    and/or openvpn and ikev2 based with certificates. To guarantee maximum
    compatibility, it is recommended to pass both objects for
    username/password and certificates.
    """
    pubkey_credentials: Optional[VPNPubkeyCredentials]
    userpass_credentials: Optional[VPNUserPassCredentials]


class Features(Protocol):
    """
    This class is used to define which features are supported.
    """
    # pylint: disable=too-few-public-methods duplicate-code
    netshield: int
    moderate_nat: bool
    vpn_accelerator: bool
    port_forwarding: bool
    ipv6: bool


class Settings(Protocol):
    """Optional.

    If you would like to pass some specific settings for VPN
    configuration then you should derive from this class and override
    its methods.

    Usage:

    .. code-block::

        from proton.vpn.connection import Settings

        class VPNSettings(Settings):

            @property
            def dns_custom_ips(self):
                return ["192.12.2.1", "175.12.3.5"]

    Note: Not all fields are mandatory to override, only those that are
    actually needed, ie:

    .. code-block::

        from proton.vpn.connection import Settings

        class VPNSettings(Settings):

            @property
            def dns_custom_ips(self):
                return ["192.12.2.1", "175.12.3.5"]

    Passing only this is perfectly fine.
    """
    # pylint: disable=too-few-public-methods
    killswitch: int
    dns_custom_ips: List[str]
    features: Features
    protocol: str



"""
Proton VPN Logging API.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from datetime import datetime, timezone
import logging
import os
from logging.handlers import RotatingFileHandler
from proton.utils.environment import VPNExecutionEnvironment


def _format_log_attributes(category, subcategory, event, optional, msg):
    """Format the log message as per Proton VPN guidelines.

        param category: Category of a log, uppercase.
        :type category: string
        param subcategory: Subcategory of a log, uppercase (optional).
        :type subcategory: string
        param event: Event of a log, uppercase.
        :type event: string
        param optional: Additional contextual data (optional).
        :type optional: string
        param msg: The message, should contain all necessary details that
            help better understand the reason behind the message.
        :type msg: string
    """
    _category = f"{category}" if category else ""
    _subcategory = f".{subcategory}" if subcategory else ""
    _event = f":{event}" if event else ""
    _optional = f" | {optional}" if optional else ""

    _msg = ""
    if msg:
        _msg = f" | {msg}" if event else f"{msg}"

    return f"{_category.upper()}{_subcategory.upper()}{_event.upper()}{_msg}{_optional}"


class ProtonAdapter(logging.LoggerAdapter):
    """Adapter to add the allowed Proton attributes"""
    ALLOWED_PROTON_ATTRS = ["category", "subcategory", "event", "optional"]

    def process(self, msg, kwargs):
        # Obtain all Proton logging attributes from kwargs.
        # Note that they should be removed from the kwargs dict as well
        # before delegating to logging.Logger. Otherwise, logging.Logger
        # would raise an error due to unrecognized kwargs.
        category = kwargs.pop("category", None)
        subcategory = kwargs.pop("subcategory", None)
        event = kwargs.pop("event", None)
        optional = kwargs.pop("optional", None)

        return _format_log_attributes(category, subcategory, event, optional, msg), kwargs


def getLogger(name):  # noqa # pylint: disable=C0103
    """
    Returns the logger with the specified name, wrapped in a
    logging.LoggerAdapter which adds the Proton attributes to the log message.

    The allowed proton attributes are: category, subcategory, event and optional.

    Usage:
    .. highlight:: python
    .. code-block:: python

        import proton.vpn.core_api.vpn_logging as logging

         # 1. config should be called asap, but only once.
        logging.config("my_log_file")

        # 2. Get a logger per module.
        logger = logging.getLogger(__name__)

        # 3. Use any of the logger methods (debug, warning, info, error, exception,..)
        # passing the allowed Proton attributes (or not).
        logger.info(
            "my message",
            category="my_category",
            subcategory="my_subcategory",
            event="my_event",
            optional="optional stuff"
        )

    The resulting log message should look like this:

    2022-09-20T07:59:27.393743 | INFO | MY_CATEGORY.MY_SUBCATEGORY:MY_EVENT
    | my message | optional stuff
    """
    return ProtonAdapter(logging.getLogger(name), extra={})


def config(filename, logdirpath=None, log_to_console=True):
    """Configure root logger.

        param filename: Log filename without extension.
        :type filename: string
        param logdirpath: Path to log file (optional).
        :type logdirpath: string
        param log_to_console: Determines whether logs will output to console (optional)
        :type log_to_console: bool
    """
    logger = logging.getLogger()
    logging_level = logging.INFO

    if filename is None:
        raise ValueError("Filename must be set")

    filename = filename + ".log"

    default_logdirpath = os.path.join(VPNExecutionEnvironment().path_cache, "logs")
    logdirpath = logdirpath or default_logdirpath

    log_filepath = os.path.join(logdirpath, filename)

    os.makedirs(logdirpath, mode=0o700, exist_ok=True)

    _formatter = logging.Formatter(
        fmt="%(asctime)s | %(name)s:%(lineno)d | %(levelname)s | %(message)s",
    )
    _formatter.formatTime = (
        lambda record, datefmt=None: datetime.now(timezone.utc).isoformat()
    )

    # Starts a new file at 3MB size limit
    _handler_file = RotatingFileHandler(
        log_filepath, maxBytes=3145728, backupCount=3
    )
    _handler_file.setFormatter(_formatter)

    if log_to_console:
        # Handler to log to console
        _handler_console = logging.StreamHandler()
        _handler_console.setFormatter(_formatter)
        logger.addHandler(_handler_console)

    # Only log debug when using PROTON_VPN_DEBUG=true
    if os.environ.get("PROTON_VPN_DEBUG", "false").lower() == "true":
        logging_level = logging.DEBUG

    logger.setLevel(logging_level)
    if _handler_file:
        logger.addHandler(_handler_file)


__all__ = ["getLogger", "config"]



"""
Proton VPN Session API.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from dataclasses import dataclass

from pathlib import Path
import platform
from typing import Optional

import distro

from proton.sso import ProtonSSO
from proton.vpn import logging
from proton.vpn.connection import VPNCredentials
from proton.vpn.session import VPNSession
from proton.vpn.session.utils import to_semver_build_metadata_format, get_core_api_semver_version

logger = logging.getLogger(__name__)

CPU_ARCHITECTURE = to_semver_build_metadata_format(platform.machine())
DISTRIBUTION_ID = distro.id()
DISTRIBUTION_VERSION = distro.version()


def _is_beta_repo_installed() -> bool:
    if distro.id() == "debian" or distro.like() == "debian":
        return Path("/etc/apt/sources.list.d/protonvpn-beta.sources").is_file()

    if distro.id() == "fedora" or distro.like() == "fedora":
        return Path("/etc/yum.repos.d/protonvpn-beta.repo").is_file()

    return False


BETA_REPO_INSTALLED = _is_beta_repo_installed()


@dataclass
class ClientTypeMetadata:  # pylint: disable=missing-class-docstring
    type: str
    version: str = get_core_api_semver_version()


class SessionHolder:
    """Holds the current session object, initializing it lazily when requested."""

    def __init__(
        self, client_type_metadata: ClientTypeMetadata,
        session: VPNSession = None
    ):
        self._proton_sso = ProtonSSO(
            appversion=self._get_app_version_header_value(client_type_metadata),
            user_agent=f"ProtonVPN/{client_type_metadata.version} "
                       f"(Linux; {DISTRIBUTION_ID}/{DISTRIBUTION_VERSION})"
        )
        self._session = session

    def get_session_for(self, username: str) -> VPNSession:
        """
        Returns the session for the specified user.
        :param username: Proton account username.
        :return:
        """
        self._session = self._proton_sso.get_session(
            account_name=username,
            override_class=VPNSession
        )
        return self._session

    @property
    def session(self) -> VPNSession:
        """Returns the current session object."""
        if not self._session:
            self._session = self._proton_sso.get_default_session(
                override_class=VPNSession
            )

        return self._session

    @property
    def user_tier(self) -> Optional[int]:
        """Returns the user tier, if the session is already loaded."""
        if self.session.loaded:
            return self.session.vpn_account.max_tier

        return None

    @property
    def vpn_credentials(self) -> Optional[VPNCredentials]:
        """Returns the VPN credentials, if the session is already loaded."""
        if self.session.loaded:
            return self.session.vpn_account.vpn_credentials

        return None

    @classmethod
    def _get_app_version_header_value(cls, client_type_metadata: ClientTypeMetadata) -> str:
        app_version = f"linux-vpn-{client_type_metadata.type}@{client_type_metadata.version}"

        version_metadata = cls._get_version_metadata()
        if version_metadata:
            app_version += f"+{version_metadata}"

        return app_version

    @staticmethod
    def _get_version_metadata():
        """
        The following version metadata is sent to the REST API for the following reasons:
        1. We want to have an idea on the amount of total users running on each CPU architecture,
           to know which archs we should target when compiling components written in rust.
        2. We want to know if users have our early release linux repositories installed, so that
           2.a We have an idea of the amount of early release users.
           2.b We can enable feature flags only for early release users (i.e. users with our
               official beta release package installed).

        The reason why we do this unconventional use of the semver build metadata section is
        that currently our infra doesn't allow passing/parsing this information through another
        header that's not x-pm-appversion.
        """
        metas = []
        if CPU_ARCHITECTURE:
            metas.append(CPU_ARCHITECTURE)
        if BETA_REPO_INSTALLED:
            metas.append("beta")

        return ".".join(metas)



"""
This module contains the exceptions to be used by kill swtich backends.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""


class KillSwitchException(Exception):
    """Base class for KillSwitch specific exceptions."""

    def __init__(self, message: str, additional_context: object = None):  # noqa
        self.message = message
        self.additional_context = additional_context
        super().__init__(self.message)


class MissingKillSwitchBackendDetails(KillSwitchException):
    """When no KillSwitch backend is found then this exception is raised.

    In rare cases where it can happen that a user has some default packages installed, where the
    services for those packages are actually not running. Ie:
    NetworkManager is installed but not running and for some reason we can't access it,
    thus this exception is raised as we can't do anything.
    """



"""
Protun protocol implementation.


Copyright (c) 2026 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
# pylint: disable=R0801 # Duplicate code with wireguard.py and openvpn.py but that's expected for now.

from __future__ import annotations

import json
import socket
import uuid
import logging
from concurrent.futures import Future
from getpass import getuser
from ipaddress import IPv4Address, IPv6Address
from typing import Dict, Optional, Union, List

import gi

gi.require_version("NM", "1.0")  # noqa: required before importing NM module
# pylint: disable=wrong-import-position
from gi.repository import NM

from proton.vpn.connection import events, states
from proton.vpn.connection.events import EventContext
from proton.vpn.connection.interfaces import Settings
from proton.vpn.backend.networkmanager.core import LinuxNetworkManager, LocalAgentMixin


logger = logging.getLogger(__name__)

SERVICE_TYPE = "org.freedesktop.NetworkManager.protun"
STORE_PRIVATE_KEY_IN_KEYRING = "1"
PRIVATE_KEY = "private-key"
PRIVATE_KEY_FLAGS = "private-key-flags"

# Internal network configuration — same WireGuard infrastructure as wireguard.py
_INTERNAL_IPV4_ADDRESS = "10.2.0.2"
_INTERNAL_IPV4_PREFIX = 32
_INTERNAL_IPV4_DNS = "10.2.0.1"
_INTERNAL_IPV4_DNS_SEARCH = "~"
_DNS_PRIORITY = -1500


class Protun(LinuxNetworkManager, LocalAgentMixin):
    """Creates a protun VPN plugin connection."""

    SIGNAL_NAME: str = "vpn-state-changed"
    VIRTUAL_DEVICE_NAME: str = "proton0"
    PLUGIN_NAME: str = "protun"
    connection: Optional[NM.SimpleConnection] = None
    plugin_exists: Optional[bool] = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        LocalAgentMixin.__init__(self)
        self._connection_settings = None

    def setup(self) -> Future:
        """Creates and registers the NM VPN connection."""
        self._generate_connection()
        self._modify_connection()
        return self.nm_client.add_connection_async(self.connection)

    def _generate_connection(self):
        self._unique_id = str(uuid.uuid4())
        self._connection_settings = NM.SettingConnection.new()
        self.connection = NM.SimpleConnection.new()

    def _modify_connection(self):
        self._set_custom_connection_id()
        self._set_uuid()
        self._set_interface_name()
        self._set_connection_type()
        self._set_connection_user_owned()
        self.connection.add_setting(self._connection_settings)

        self._set_route()
        self._set_dns()
        self._set_vpn_settings()

        self.connection.verify()

    def _set_custom_connection_id(self):
        self._connection_settings.set_property(NM.SETTING_CONNECTION_ID, self._get_servername())

    def _set_uuid(self):
        self._connection_settings.set_property(NM.SETTING_CONNECTION_UUID, self._unique_id)

    def _set_interface_name(self):
        self._connection_settings.set_property(
            NM.SETTING_CONNECTION_INTERFACE_NAME, self.VIRTUAL_DEVICE_NAME
        )

    def _set_connection_type(self):
        self._connection_settings.set_property(
            NM.SETTING_CONNECTION_TYPE, NM.SETTING_VPN_SETTING_NAME
        )

    def _set_connection_user_owned(self):
        self._connection_settings.add_permission(
            NM.SETTING_USER_SETTING_NAME,
            getuser(),
            None
        )

    def _set_route(self):
        ipv4_config = NM.SettingIP4Config.new()
        ipv6_config = NM.SettingIP6Config.new()

        # We set SETTING_IP4_CONFIG_METHOD_AUTO so that the network manager
        # sets up automatic routing to the VPN server IP, to avoid routing
        # loops.
        #
        # Later we'll want to replace this with fwmark which will allow protun
        # to work with split tunneling.
        ipv4_config.set_property(NM.SETTING_IP_CONFIG_METHOD,
                                 NM.SETTING_IP4_CONFIG_METHOD_AUTO)
        ipv4_config.add_address(
            NM.IPAddress.new(socket.AF_INET, _INTERNAL_IPV4_ADDRESS, _INTERNAL_IPV4_PREFIX)
        )

        # IPv6 is not yet supported by the protun plugin
        ipv6_config.set_property(
            NM.SETTING_IP_CONFIG_METHOD, NM.SETTING_IP6_CONFIG_METHOD_DISABLED
        )

        self.connection.add_setting(ipv4_config)
        self.connection.add_setting(ipv6_config)

    def _set_dns(self):
        ipv4_config = self.connection.get_setting_ip4_config()
        self._configure_dns(nm_setting=ipv4_config, ip_version=IPv4Address)
        self.connection.add_setting(ipv4_config)

    def _configure_dns(
        self,
        nm_setting: Union[NM.SettingIP4Config, NM.SettingIP6Config],
        ip_version: Union[IPv4Address, IPv6Address],
        dns_priority: int = _DNS_PRIORITY,
    ):
        """Sets DNS, respecting custom DNS settings."""
        if ip_version not in [IPv4Address, IPv6Address]:
            raise ValueError(f"Unknown IP version: {ip_version}")

        nm_setting.set_property(NM.SETTING_IP_CONFIG_DNS_PRIORITY, dns_priority)
        nm_setting.set_property(NM.SETTING_IP_CONFIG_IGNORE_AUTO_DNS, True)

        # pylint: disable=duplicate-code
        custom_dns_ips = self._settings.custom_dns\
            .get_enabled_dns_list_based_on_ip_version(ip_version)
        ip_addresses = [dns.exploded for dns in custom_dns_ips]

        # Protun does not auto-configure DNS, so we always set it explicitly.
        if self._settings.custom_dns.enabled and ip_addresses:
            nm_setting.set_property(NM.SETTING_IP_CONFIG_DNS, ip_addresses)
        else:
            nm_setting.add_dns(_INTERNAL_IPV4_DNS)
            nm_setting.add_dns_search(_INTERNAL_IPV4_DNS_SEARCH)

    def _protun_ports(self) -> Dict[str, List[int]]:
        """Returns the protun ports as a dict."""
        raise NotImplementedError(
            "This method should be implemented by ProtunUDP, ProtunTCP, etc. subclasses.")

    def _set_vpn_settings(self):
        vpn_settings = NM.SettingVpn.new()
        vpn_settings.set_property(NM.SETTING_VPN_SERVICE_TYPE, SERVICE_TYPE)

        # Build the peers array expected by the protun plugin (settings.rs)
        peer = {
            "id": self._vpnserver.server_name or self._vpnserver.server_ip,
            "endpoint": f"{self._vpnserver.server_ip}",
            "public-key": self._vpnserver.x25519pk,
            "udp-ports": [],
            "tcp-ports": [],
            "tls-ports": [],
            "priority": 0,
        }
        peer.update(self._protun_ports())
        peers = [peer]
        vpn_settings.add_data_item("peers", json.dumps(peers))

        # The WireGuard private key is stored as a VPN secret.
        # NM passes it to the protun auth-dialog which forwards it to the plugin.
        vpn_settings.add_secret(
            PRIVATE_KEY,
            self._vpncredentials.pubkey_credentials.wg_private_key
        )

        # Use the keyring to store the connection private key.
        vpn_settings.add_data_item(PRIVATE_KEY_FLAGS,
                                   STORE_PRIVATE_KEY_IN_KEYRING)

        self.connection.add_setting(vpn_settings)

    # pylint: disable=arguments-renamed
    def _on_state_changed(
            self, vpn_connection: NM.VpnConnection, state: int, reason: int
    ):
        """
        Handle VPN plugin state changes emitted by NetworkManager.

        NM emits vpn-state-changed with state and reason whenever the protun
        plugin connection state changes. This translates NM states into
        backend-agnostic state machine events.
        """
        try:
            state = NM.VpnConnectionState(state)
        except ValueError:
            logger.warning("Unexpected VPN connection state: %s", state)
            state = NM.VpnConnectionState.UNKNOWN

        try:
            reason = NM.VpnConnectionStateReason(reason)
        except ValueError:
            logger.warning("Unexpected VPN connection state reason: %s", reason)
            reason = NM.VpnConnectionStateReason.UNKNOWN

        logger.debug(
            "Protun connection state changed: state=%s, reason=%s",
            state.value_name, reason.value_name
        )

        if state is NM.VpnConnectionState.ACTIVATED:
            logger.info("Starting local agent for domain: %s", self._vpnserver.domain)
            self._async_start_local_agent_listener()
            self._notify_subscribers_threadsafe(
                events.Connected(EventContext(connection=self))
            )
        elif state is NM.VpnConnectionState.FAILED:
            if reason in [
                NM.VpnConnectionStateReason.CONNECT_TIMEOUT,
                NM.VpnConnectionStateReason.SERVICE_START_TIMEOUT,
            ]:
                self._notify_subscribers_threadsafe(
                    events.Timeout(EventContext(connection=self, reason=reason))
                )
            elif reason in [
                NM.VpnConnectionStateReason.NO_SECRETS,
                NM.VpnConnectionStateReason.LOGIN_FAILED,
            ]:
                self._notify_subscribers_threadsafe(
                    events.AuthDenied(EventContext(connection=self, reason=reason))
                )
            else:
                self._notify_subscribers_threadsafe(
                    events.UnexpectedError(EventContext(connection=self, reason=reason))
                )
        elif state is NM.VpnConnectionState.DISCONNECTED:
            if reason is NM.VpnConnectionStateReason.USER_DISCONNECTED:
                self._async_stop_local_agent_listener()
                self._notify_subscribers_threadsafe(
                    events.Disconnected(EventContext(connection=self, reason=reason))
                )
            elif reason is NM.VpnConnectionStateReason.DEVICE_DISCONNECTED:
                self._async_stop_local_agent_listener()
                self._notify_subscribers_threadsafe(
                    events.DeviceDisconnected(EventContext(connection=self, reason=reason))
                )
            else:
                self._notify_subscribers_threadsafe(
                    events.UnexpectedError(EventContext(connection=self, reason=reason))
                )
        else:
            logger.debug("Ignoring VPN state change: %s", state.value_name)

    def _initialize_persisted_connection(
            self, connection_id: str
    ) -> states.State:
        """Initialize from a persisted connection, restarting the local agent if connected."""
        state = super()._initialize_persisted_connection(connection_id)

        if isinstance(state, states.Connected):
            self._async_start_local_agent_listener()
        return state

    async def update_credentials(self, credentials):
        """Notifies the VPN server that the certificate needs a refresh."""
        await super().update_credentials(credentials)
        await self._start_local_agent_listener()

    @property
    def are_feature_updates_applied_when_active(self) -> bool:
        """
        Returns whether connection feature updates are applied on the fly
        while the connection is already active, without restarting it.
        """
        return True

    async def update_settings(self, settings: Settings):
        """Update features on the active agent connection."""
        await super().update_settings(settings)
        if self._agent_listener.is_running:  # noqa: E501 # pylint: disable=line-too-long # nosemgrep: python.lang.maintainability.is-function-without-parentheses.is-function-without-parentheses
            await self._request_connection_features(settings.features)

    @classmethod
    def _get_priority(cls):
        return 1

    @classmethod
    def _validate(cls):
        if cls.plugin_exists is None:
            cls.plugin_exists = cls._plugin_exists(cls.PLUGIN_NAME)
        return cls.plugin_exists


class ProtunUDP(Protun):
    """Protun UDP protocol implementation."""

    protocol = "protun-udp"
    ui_protocol = "Protun (UDP)"

    def _protun_ports(self) -> Dict[str, List[int]]:
        """Returns the protun ports as a dict."""
        return {
            "udp-ports": self._vpnserver.wireguard_ports.udp
        }


class ProtunTCP(Protun):
    """Protun TCP protocol implementation."""

    protocol = "protun-tcp"
    ui_protocol = "Protun (TCP)"

    def _protun_ports(self) -> Dict[str, List[int]]:
        """Returns the protun ports as a dict."""
        return {
            "tcp-ports": self._vpnserver.wireguard_ports.tcp
        }


class ProtunTLS(Protun):
    """Protun TLS protocol implementation."""

    protocol = "protun-tls"
    ui_protocol = "Protun (Stealth)"

    def _protun_ports(self) -> Dict[str, List[int]]:
        """Returns the protun ports as a dict."""
        return {
            "tls-ports": self._vpnserver.wireguard_ports.tls
        }



"""
Module that contains the base class for Kill Switch implementations to extend from.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations

from abc import ABC, abstractmethod
from enum import IntEnum
from typing import TYPE_CHECKING, Optional

from proton.loader import Loader

from proton.vpn.killswitch.interface.exceptions import MissingKillSwitchBackendDetails

if TYPE_CHECKING:
    from proton.vpn.connection import VPNServer


class KillSwitchState(IntEnum):  # pylint: disable=missing-class-docstring
    OFF = 0
    ON = 1
    PERMANENT = 2


class KillSwitch(ABC):
    """
    The `KillSwitch` is the base class from which all other kill switch
    backends need to derive from.
    """
    @staticmethod
    def get(class_name: str = None, protocol: str = None) -> KillSwitch:
        """
        Returns the kill switch implementation.

        :param class_name: Name of the class implementing the kill switch. This
        parameter is optional. If it's not provided then the existing implementation
        with the highest priority is returned.
        :param protocol: the kill switch backend to be used based on protocol.
            This is mainly used for backend validation.
        """
        try:
            return Loader.get(
                type_name="killswitch",
                class_name=class_name,
                validate_params={"protocol": protocol}
            )
        except RuntimeError as excp:
            raise MissingKillSwitchBackendDetails(excp) from excp

    @abstractmethod
    async def enable(self, vpn_server: Optional["VPNServer"] = None, permanent: bool = False):
        """
        Enables the kill switch.
        """

    @abstractmethod
    async def disable(self):
        """
        Disables the kill switch.
        """

    @abstractmethod
    async def enable_ipv6_leak_protection(self, permanent: bool = False):
        """
        Enables IPv6 kill switch to prevent leaks.
        """

    @abstractmethod
    async def disable_ipv6_leak_protection(self):
        """
        Disables IPv6 kill switch to prevent leaks.
        """

    @staticmethod
    @abstractmethod
    def _get_priority() -> int:
        pass

    @staticmethod
    @abstractmethod
    def _validate():
        pass



"""
This module manages the Proton VPN general settings.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from dataclasses import dataclass
from enum import IntEnum

from proton.vpn.core.settings.split_tunneling import SplitTunneling

from proton.vpn import logging

logger = logging.getLogger(__name__)


class NetShield(IntEnum):  # pylint: disable=missing-class-docstring
    NO_BLOCK = 0
    BLOCK_MALICIOUS_URL = 1
    BLOCK_ADS_AND_TRACKING = 2


@dataclass
class Features:
    """Contains features that affect a vpn connection"""
    # pylint: disable=duplicate-code
    netshield: int
    moderate_nat: bool
    vpn_accelerator: bool
    port_forwarding: bool
    split_tunneling: SplitTunneling

    def are_free_tier_defaults(self):
        """Returns True if the features are the ones set for the free tier."""
        return self == Features.default(user_tier=0)

    @staticmethod
    def from_dict(data: dict, user_tier: int) -> Features:
        """Creates and returns `Features` from the provided dict."""
        default = Features.default(user_tier)
        split_tunneling = data.get("split_tunneling")
        split_tunneling = SplitTunneling.from_dict(split_tunneling) \
            if split_tunneling else SplitTunneling()

        return Features(
            netshield=data.get("netshield", default.netshield),
            moderate_nat=data.get("moderate_nat", default.moderate_nat),
            vpn_accelerator=data.get(
                "vpn_accelerator", default.vpn_accelerator),
            port_forwarding=data.get(
                "port_forwarding", default.port_forwarding),
            split_tunneling=split_tunneling
        )

    def to_dict(self) -> dict:
        """Converts the class to dict."""
        return {
            "netshield": self.netshield,
            "moderate_nat": self.moderate_nat,
            "vpn_accelerator": self.vpn_accelerator,
            "port_forwarding": self.port_forwarding,
            "split_tunneling": self.split_tunneling.to_dict(),
        }

    @staticmethod
    def default(user_tier: int) -> Features:  # pylint: disable=unused-argument
        """Creates and returns `Features` from default configurations."""
        return Features(
            netshield=(
                NetShield.NO_BLOCK.value
                if user_tier < 1
                else NetShield.BLOCK_MALICIOUS_URL.value
            ),
            moderate_nat=False,
            vpn_accelerator=True,
            port_forwarding=False,
            split_tunneling=SplitTunneling()
        )

    def is_default(self, user_tier: int) -> bool:
        """Returns true if the features are the default ones."""
        return self == Features.default(user_tier)



"""
The different VPN connection states and their transitions is defined here.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Optional, ClassVar


from proton.vpn import logging
from proton.vpn.connection import events
from proton.vpn.connection.enum import ConnectionStateEnum, KillSwitchSetting
from proton.vpn.connection.events import EventContext
from proton.vpn.connection.exceptions import ConcurrentConnectionsError
from proton.vpn.killswitch.interface import KillSwitch
from proton.vpn.split_tunneling.exceptions import SplitTunnelingError
from proton.vpn.split_tunneling.interface import SplitTunneling
from proton.vpn.core.settings.split_tunneling import SplitTunneling as SplitTunnelingSetting


if TYPE_CHECKING:
    from proton.vpn.connection.vpnconnection import VPNConnection


logger = logging.getLogger(__name__)


@dataclass
class StateContext:
    """
    Relevant state context data.

    Attributes:
        event: Event that led to the current state.
        connection: current VPN connection. They only case where this
            attribute could be None is on the initial state, if there is not
            already an existing VPN connection.
        reconnection: optional VPN connection to connect to as soon as stopping the current one.
        kill_switch: kill switch implementation.
        kill_switch_setting: on, off, permanent.
        split_tunneling: split tunneling implementation, if available.
        split_tunneling_setting: split tunneling configuration.
    """
    event: events.Event = field(default_factory=events.Initialized)
    connection: Optional["VPNConnection"] = None
    reconnection: Optional["VPNConnection"] = None
    kill_switch: ClassVar[KillSwitch] = None
    kill_switch_setting: ClassVar[KillSwitchSetting] = None
    split_tunneling: ClassVar[Optional[SplitTunneling]] = None
    split_tunneling_setting: ClassVar[SplitTunnelingSetting] = None


class State(ABC):
    """
    This is the base state from which all other states derive from. Each new
    state has to implement the `on_event` method.

    Since these states are backend agnostic. When implement a new backend the
    person implementing it has to have special care in correctly translating
    the backend specific events to known events
    (see `proton.vpn.connection.events`).

    Each state acts on the `on_event` method. Generally, if a state receives
    an unexpected event, it will then not update the state but rather keep the
    same state and should log the occurrence.

    The general idea of state transitions:

        1) Connect happy path:      Disconnected -> Connecting -> Connected
        2) Connect with error path: Disconnected -> Connecting -> Error
        3) Disconnect happy path:   Connected -> Disconnecting -> Disconnected
        4) Active connection error path: Connected -> Error

    Certain states will have to call methods from the state machine
    (see `Disconnected`, `Connected`). Both of these states call
    `vpn_connection.start()` and `vpn_connection.stop()`.
    It should be noted that these methods should be run in an async way so that
    it does not block the execution of the next line.

    States also have `context` (which are fetched from events). These can help
    in discovering potential issues on why certain states might an unexpected
    behavior. It is worth mentioning though that the contexts will always
    be backend specific.
    """
    type = None

    # Indicates whether this state should be notified early (i.e. before running
    # the tasks associated to the state) or not (i.e. after running the tasks
    # associated to the state)
    notify_early = False

    def __init__(self, context: StateContext = None):
        self.context = context or StateContext()

        if self.type is None:
            raise TypeError("Undefined attribute \"state\" ")

    def _assert_no_concurrent_connections(self, event: events.Event):
        not_up_event = not isinstance(event, events.Up)
        different_connection = event.context.connection is not self.context.connection
        if not_up_event and different_connection:
            # Any state should always receive events for the same connection, the only
            # exception being when the Up event is received. In this case, the Up event
            # always carries a new connection: the new connection to be initiated.
            raise ConcurrentConnectionsError(
                f"State {self} expected events from {self.context.connection} "
                f"but received an event from {event.context.connection} instead."
            )

    def on_event(self, event: events.Event) -> State:
        """Returns the new state based on the received event."""
        self._assert_no_concurrent_connections(event)

        event.check_for_errors()

        new_state = self._on_event(event)

        if new_state is self:
            logger.warning(
                f"{self.type.name} state received unexpected "
                f"event: {type(event).__name__}",
                category="CONN", event="WARNING"
            )

        return new_state

    @abstractmethod
    def _on_event(
            self, event: events.Event
    ) -> State:
        """Given an event, it returns the new state."""

    async def run_tasks(self) -> Optional[events.Event]:
        """Tasks to be run when this state instance becomes the current VPN state."""

    @property
    def forwarded_port(self) -> Optional[int]:
        """Returns the forwarded port if it exists."""
        return self.context.event.context.forwarded_port


class Disconnected(State):
    """
    Disconnected is the initial state of a connection. It's also its final
    state, except if the connection could not be established due to an error.
    """
    type = ConnectionStateEnum.DISCONNECTED

    def _on_event(self, event: events.Event):
        if isinstance(event, events.Up):
            return Connecting(StateContext(event=event, connection=event.context.connection))

        return self

    async def run_tasks(self):
        # When the state machine is in disconnected state, a VPN connection
        # may have not been created yet.
        if self.context.connection:
            await self.context.connection.remove_persistence()

        if self.context.reconnection:
            # The Kill switch is enabled to avoid leaks when switching servers, even when
            # the kill switch setting is off.
            await self.context.kill_switch.enable()

            # When a reconnection is expected, an Up event is returned to start a new connection.
            # straight away.
            return events.Up(EventContext(connection=self.context.reconnection))

        if self.context.kill_switch_setting == KillSwitchSetting.PERMANENT:
            # This is an abstraction leak of the network manager KS.
            # The only reason for enabling permanent KS here is to switch from the
            # routed KS to the full KS if the user cancels the connection while in
            # Connecting state. Otherwise, the full KS should already be there.
            await self.context.kill_switch.enable(permanent=True)
        else:
            await self.context.kill_switch.disable()
            await self.context.kill_switch.disable_ipv6_leak_protection()

        if self.context.split_tunneling:
            # ST config is always cleared independently of if ST is disabled via settings or
            # via feature flag to make sure existing config is always cleaned up.
            try:
                await self.context.split_tunneling.clear_config()
            except SplitTunnelingError:
                # We decided to treat split tunneling error as non-fatal, to prevent they
                # impact the core VPN functionality.
                logger.exception("Error clearing split tunneling configuration")

        return None


class Connecting(State):
    """
    Connecting is the state reached when a VPN connection is requested.
    """
    type = ConnectionStateEnum.CONNECTING
    # Connection state subscribers will be notified of the Connecting state before running
    # the tasks associated to this state.
    notify_early = True

    def _on_event(self, event: events.Event):
        if isinstance(event, events.Connected):
            return Connected(StateContext(event=event, connection=event.context.connection))

        if isinstance(event, events.Down):
            return Disconnecting(StateContext(event=event, connection=event.context.connection))

        if isinstance(event, events.Error):
            return Error(StateContext(event=event, connection=event.context.connection))

        if isinstance(event, events.Up):
            # If a new connection is requested while in `Connecting` state then
            # cancel the current one and pass the requested connection so that it's
            # started as soon as the current connection is down.
            return Disconnecting(
                StateContext(
                    event=event,
                    connection=self.context.connection,
                    reconnection=event.context.connection
                )
            )

        if isinstance(event, events.Disconnected):
            # Another process disconnected the VPN, otherwise the Disconnected
            # event would've been received by the Disconnecting state.
            return Disconnected(StateContext(event=event, connection=event.context.connection))

        return self

    async def run_tasks(self):
        permanent_ks = self.context.kill_switch_setting == KillSwitchSetting.PERMANENT

        # The reason for always enabling the kill switch independently of the kill switch setting
        # is to avoid leaks when switching servers, even with the kill switch turned off.
        # However, when the kill switch setting is off, the kill switch has to be removed when
        # reaching the connected state.
        await self.context.kill_switch.enable(
            self.context.connection.server,
            permanent=permanent_ks
        )

        await self.context.connection.start()


class Connected(State):
    """
    Connected is the state reached once the VPN connection has been successfully
    established.
    """
    type = ConnectionStateEnum.CONNECTED

    def _on_event(self, event: events.Event):
        if isinstance(event, events.Down):
            return Disconnecting(StateContext(event=event, connection=event.context.connection))

        if isinstance(event, events.Up):
            # If a new connection is requested while in `Connected` state then
            # cancel the current one and pass the requested connection so that it's
            # started as soon as the current connection is down.
            return Disconnecting(
                StateContext(
                    event=event,
                    connection=self.context.connection,
                    reconnection=event.context.connection
                )
            )

        if isinstance(event, events.Error):
            return Error(StateContext(event=event, connection=event.context.connection))

        if isinstance(event, events.Disconnected):
            # Another process disconnected the VPN, otherwise the Disconnected
            # event would've been received by the Disconnecting state.
            return Disconnected(StateContext(event=event, connection=event.context.connection))

        if isinstance(event, events.Connected):
            return Connected(StateContext(event=event, connection=event.context.connection))

        return self

    async def run_tasks(self):
        if self.context.kill_switch_setting == KillSwitchSetting.OFF:
            await self.context.kill_switch.enable_ipv6_leak_protection()
            await self.context.kill_switch.disable()
            if self.context.split_tunneling_setting.enabled:
                try:
                    await self.context.split_tunneling.set_config(
                        self.context.split_tunneling_setting
                            .get_config()
                    )
                except SplitTunnelingError:
                    # We decided to treat split tunneling error as non-fatal, to prevent they
                    # impact the core VPN functionality.
                    logger.exception("Error setting split tunnel configuration")

        else:
            # This is specific to the routing table KS implementation and should be removed.
            # At this point we switch from the routed KS to the full-on KS.
            await self.context.kill_switch.enable(
                permanent=(self.context.kill_switch_setting == KillSwitchSetting.PERMANENT)
            )

        await self.context.connection.add_persistence()


class Disconnecting(State):
    """
    Disconnecting is state reached when VPN disconnection is requested.
    """
    type = ConnectionStateEnum.DISCONNECTING
    # Connection state subscribers will be notified of the Disconnecting state before running
    # the tasks associated to this state.
    notify_early = True

    def _on_event(self, event: events.Event):
        if isinstance(event, (events.Disconnected, events.Error)):
            # Note that error events signal disconnection from the VPN due to
            # unexpected reasons. In this case, since the goal of the
            # disconnecting state is to reach the disconnected state,
            # both disconnected and error events lead to the desired state.
            if isinstance(event, events.Error):
                logger.warning(
                    "Error event while disconnecting: %s (%s)",
                    type(event).__name__,
                    event.context.error
                )
            return Disconnected(
                StateContext(
                    event=event,
                    connection=event.context.connection,
                    reconnection=self.context.reconnection
                )
            )

        if isinstance(event, events.Up):
            # If a new connection is requested while in the `Disconnecting` state then
            # store the requested connection in the state context so that it's started
            # as soon as the current connection is down.
            self.context.reconnection = event.context.connection

        return self

    async def run_tasks(self):
        await self.context.connection.stop()


class Error(State):
    """
    Error is the state reached after a connection error.
    """
    type = ConnectionStateEnum.ERROR

    def _on_event(self, event: events.Event):
        if isinstance(event, events.Down):
            return Disconnected(StateContext(event=event, connection=event.context.connection))

        if isinstance(event, events.Up):
            return Disconnecting(
                StateContext(
                    event=event,
                    connection=self.context.connection,
                    reconnection=event.context.connection
                )
            )

        if isinstance(event, events.Connected):
            return Connected(
                StateContext(
                    event=event,
                    connection=self.context.connection,
                )
            )

        if isinstance(event, events.Error):
            return Error(StateContext(event=event, connection=event.context.connection))

        return self

    async def run_tasks(self):

        logger.warning(
            "Reached connection error state: %s (%s)",
            type(self.context.event).__name__,
            self.context.event.context.error
        )



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from typing import List, IO
from dataclasses import dataclass, field
from proton.vpn.session.utils import generate_os_string, get_distro_version

VPN_CLIENT_TYPE = "2"  # 1: email;  2: VPN

# pylint: disable=invalid-name


@dataclass
class BugReportForm:  # pylint: disable=too-many-instance-attributes
    """Bug report form data to be submitted to customer support."""
    username: str
    email: str
    title: str
    description: str
    client_version: str
    client: str
    attachments: List[IO] = field(default_factory=list)
    os: str = generate_os_string()  # pylint: disable=invalid-name
    os_version: str = get_distro_version()
    client_type: str = VPN_CLIENT_TYPE



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from datetime import timedelta
from typing import Callable, Optional

from proton.session.exceptions import (
    ProtonAPINotReachable, ProtonAPINotAvailable,
    ProtonAPIError
)

from proton.vpn import logging
from proton.vpn.core.refresher.scheduler import RunAgain
from proton.vpn.core.session_holder import SessionHolder
from proton.vpn.session.servers.logicals import ServerList

logger = logging.getLogger(__name__)


class ServerListRefresher:
    """
    Service in charge of refreshing the VPN server list/loads.
    """
    def __init__(self, session_holder: SessionHolder):
        self._session_holder = session_holder
        self.server_list_updated_callback: Optional[Callable] = None
        self.server_loads_updated_callback: Optional[Callable] = None

    @property
    def _session(self):
        return self._session_holder.session

    @property
    def initial_refresh_delay(self):
        """Returns the initial delay before the first refresh."""
        return self._session.server_list.seconds_until_expiration

    async def update_if_necessary(self) -> float:
        """Refreshes the server list/loads if expired, and returns seconds till next expiration"""
        try:
            if self._session.server_list.expired:
                server_list = await self._session.fetch_server_list()
                self._notify_server_list()
                next_refresh_delay = server_list.seconds_until_expiration
            elif self._session.server_list.loads_expired:
                server_list = await self._session.update_server_loads()
                self._notify_server_loads()
                next_refresh_delay = server_list.seconds_until_expiration
            else:
                next_refresh_delay = self._session.server_list.seconds_until_expiration
        except ProtonAPIError as error:
            if error.http_code != 429:
                raise

            logger.warning(f"Server list refresh failed: {error}")
            next_refresh_delay = ServerList.get_loads_refresh_interval_in_seconds()
        except (ProtonAPINotReachable, ProtonAPINotAvailable) as error:
            logger.warning(f"Server list refresh failed: {error}")
            next_refresh_delay = ServerList.get_loads_refresh_interval_in_seconds()
        except Exception:
            logger.error(  # noqa: E501 # pylint: disable=line-too-long # nosemgrep: python.lang.best-practice.logging-error-without-handling.logging-error-without-handling
                "Server list refresh failed unexpectedly. "
                "Stopping server list refresh."
            )
            raise

        return next_refresh_delay

    async def refresh(self) -> RunAgain:
        """Refreshes the server list/loads if expired, else schedules a future refresh."""
        next_refresh_delay = await self.update_if_necessary()

        # Let the scheduler know that this method should be run again after a delay.
        logger.info(
            f"Next server list refresh scheduled in "
            f"{timedelta(seconds=next_refresh_delay)}"
        )

        return RunAgain.after_seconds(next_refresh_delay)

    def _notify_server_loads(self):
        if callable(self.server_loads_updated_callback):
            self.server_loads_updated_callback()  # pylint: disable=not-callable

    def _notify_server_list(self):
        if callable(self.server_list_updated_callback):
            self.server_list_updated_callback()  # pylint: disable=not-callable



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations

import time
from dataclasses import dataclass
from proton.vpn.session.utils import Serializable

# pylint: disable=invalid-name


@dataclass
class VPNCertificate(Serializable):  # pylint: disable=too-many-instance-attributes
    """ Same object structure coming from the API """
    SerialNumber: str
    ClientKeyFingerprint: str
    ClientKey: str
    """ Client public key used to ask for this certificate in PEM format. """
    Certificate: str
    """ Certificate value in PEM format. Contains the features requested at fetch time"""
    ExpirationTime: int
    RefreshTime: int
    Mode: str
    DeviceName: str
    ServerPublicKeyMode: str
    ServerPublicKey: str

    @property
    def remaining_time_to_next_refresh(self) -> int:
        """Returns a timestamp of when the next refresh should be done."""
        remaining_time = self.RefreshTime - time.time()
        return remaining_time if remaining_time > 0 else 0

    @staticmethod
    def _deserialize(dict_data: dict) -> VPNCertificate:
        return VPNCertificate(
            SerialNumber=dict_data["SerialNumber"],
            ClientKeyFingerprint=dict_data["ClientKeyFingerprint"],
            ClientKey=dict_data["ClientKey"],
            Certificate=dict_data["Certificate"],
            ExpirationTime=dict_data["ExpirationTime"],
            RefreshTime=dict_data["RefreshTime"],
            Mode=dict_data["Mode"],
            DeviceName=dict_data["DeviceName"],
            ServerPublicKeyMode=dict_data["ServerPublicKeyMode"],
            ServerPublicKey=dict_data["ServerPublicKey"]
        )



"""
Base class for VPN connections created with NetworkManager.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import asyncio
import logging
import random

from concurrent import futures

from proton.vpn.connection import events
from proton.vpn.connection.events import EventContext
from proton.vpn.connection.exceptions \
    import FeatureError, FeaturePolicyError, FeatureSyntaxError
from proton.vpn.core.settings.features import Features

from proton.vpn.backend.networkmanager.core.local_agent import (
    AgentListener, Status, ExpiredCertificateError, AgentFeatures,
    ReasonCode, PolicyAPIError, SyntaxAPIError, APIError, State
)

logger = logging.getLogger(__name__)

TWOFA_REASON_CODES = [
    ReasonCode.REASON_CODE_2FA_UNSPECIFIED,
    ReasonCode.REASON_CODE_2FA_EXPIRED,
    ReasonCode.REASON_CODE_2FA_SITUATION_CHANGED
]


class LocalAgentMixin:  # pylint: disable=too-few-public-methods
    """
    This mixin provides the functionality to start and stop the local agent
    client and to listen for status updates and errors from the local agent
    server.
    """
    def __init__(self):
        self._agent_listener = AgentListener(
            on_status=self.__on_local_agent_status,
            on_error=self.__on_local_agent_error
        )

    async def _start_local_agent_listener(self):
        """
        This method starts the local agent listener and waits for it to
        complete.
        """
        if self._agent_listener.is_running:  # noqa: E501 # pylint: disable=line-too-long # nosemgrep: python.lang.maintainability.is-function-without-parentheses.is-function-without-parentheses
            logger.info("Closing existing agent connection...")
            await self._agent_listener.stop()

        logger.info("Waiting for agent status from %s...",
                    self._vpnserver.domain)
        context = EventContext(connection=self)

        agent_connection_drops = 0
        while True:
            if not await self.__attempt_to_connect_to_listener(context):
                return

            if not await self.__attempt_to_request_connection_features(context):
                return

            if not await self.__attempt_to_listen(context):
                agent_connection_drops += 1
                # nosemgrep: gitlab.bandit.B311
                sleep_seconds = random.uniform(0, 10)  # nosec B311
                logger.warning(
                    "Agent connection dropped (#%s). Retrying in %.1f seconds.",
                    agent_connection_drops, sleep_seconds
                )
                await asyncio.sleep(sleep_seconds)
                continue

            return

    # Absorb CancelledError before it is logged by Future's done callback handler
    def _handle_future_result(self, future: futures.Future):
        try:
            future.result()
        except futures.CancelledError:
            pass

    def _async_start_local_agent_listener(self):
        """
        This schedules the local agent listener to start, but it will not wait
        for it.
        """
        future = asyncio.run_coroutine_threadsafe(
            self._start_local_agent_listener(),
            self._asyncio_loop
        )
        future.add_done_callback(self._handle_future_result)

    def _async_stop_local_agent_listener(self):
        """
        This schedules the local agent listener to stop, but it will not wait
        for it.
        """
        future = asyncio.run_coroutine_threadsafe(
            self._agent_listener.stop(),
            self._asyncio_loop
        )
        future.add_done_callback(self._handle_future_result)

    async def _request_connection_features(self, features: Features):
        if features.are_free_tier_defaults():
            # No need to request default features, since they are available by default.
            # Also, the local agent server raises an error if features are requested
            # for free users, even when the requested features are the default ones.
            logger.info("Using default VPN connection features.")
            return

        logger.info("Requesting VPN connection features...")
        agent_features = self.__get_agent_features(features)
        await self._agent_listener.request_features(agent_features)
        logger.info("VPN connection features requested.")

    def __on_local_agent_status(self, status: Status):
        """The local agent listener calls this method whenever a new status is
        read from the local agent connection."""
        logger.info("Agent status received: %s", status)

        context = EventContext(
            connection=self,
            connection_details=status.connection_details,
            forwarded_port=status.features.forwarded_port if status.features else None
        )

        if status.state == State.CONNECTED:
            self._notify_subscribers_threadsafe(events.Connected(context))

        elif status.state == State.HARD_JAILED:
            self.__handle_hard_jailed_state(status)

    def __handle_hard_jailed_state(self, status: Status):
        context = EventContext(connection=self,
                               connection_details=status.connection_details)
        if status.reason.code == ReasonCode.CERTIFICATE_EXPIRED:
            self._notify_subscribers_threadsafe(
                events.ExpiredCertificate(context)
            )
        elif status.reason.code in TWOFA_REASON_CODES:
            self._notify_subscribers_threadsafe(
                events.TwoFARequired(context)
            )
        elif self.__has_reached_max_amount_of_concurrent_vpn_connections(status.reason.code):
            self._notify_subscribers_threadsafe(
                events.MaximumSessionsReached(context)
            )
        else:
            self._notify_subscribers_threadsafe(
                events.UnexpectedError(context)
            )

    def __has_reached_max_amount_of_concurrent_vpn_connections(
            self, code: ReasonCode) -> bool:
        """Check if a user has reached the maximum number of concurrent VPN
        sessions/connections permitted for the current tier."""
        return code in (
            ReasonCode.MAX_SESSIONS_UNKNOWN,
            ReasonCode.MAX_SESSIONS_FREE,
            ReasonCode.MAX_SESSIONS_BASIC,
            ReasonCode.MAX_SESSIONS_PLUS,
            ReasonCode.MAX_SESSIONS_VISIONARY,
            ReasonCode.MAX_SESSIONS_PRO
        )

    def __on_local_agent_error(self, error: Exception):
        """
        The local agent listener calls this method whenever a new error message
        read from the local agent connection.

        :param error: The error received from the local agent.
        """
        event = self.__get_event_from_error_message(error)
        self._notify_subscribers_threadsafe(event)

    def __get_event_from_error_message(self, error: Exception) -> events.Event:
        exception_message = str(error)

        if isinstance(error, PolicyAPIError):
            new_error = FeaturePolicyError(exception_message)
        elif isinstance(error, SyntaxAPIError):
            new_error = FeatureSyntaxError(exception_message)
        elif isinstance(error, APIError):
            new_error = FeatureError(exception_message)
        else:
            new_error = Exception(exception_message)

        return events.UnexpectedError(EventContext(error=new_error,
                                                   connection=self))

    async def __attempt_to_connect_to_listener(self,
                                               context: EventContext) -> bool:
        try:
            await self._agent_listener.connect(
                self._vpnserver.domain, self._vpncredentials.pubkey_credentials
            )
        except ExpiredCertificateError:
            self._notify_subscribers_threadsafe(
                events.ExpiredCertificate(context))
            return False
        except TimeoutError:
            logger.info("Connect timeout")
            self._notify_subscribers_threadsafe(events.Timeout(context))
            return False
        except Exception:
            self._notify_subscribers_threadsafe(
                events.UnexpectedError(context))
            raise

        return True

    async def __attempt_to_request_connection_features(self,
                                                       context: EventContext) -> bool:
        try:
            await self._request_connection_features(self.settings.features)
        except TimeoutError:
            self._notify_subscribers_threadsafe(events.Timeout(context))
            return False
        except Exception:
            self._notify_subscribers_threadsafe(
                events.UnexpectedError(context))
            raise

        return True

    async def __attempt_to_listen(self, context: EventContext) -> bool:
        try:
            await self._agent_listener.listen()
            return True
        except TimeoutError:
            return False
        except Exception:
            self._notify_subscribers_threadsafe(
                events.UnexpectedError(context))
            raise

    def __get_agent_features(self, features: Features) -> AgentFeatures:
        randomized_nat = (not features.moderate_nat
                          if features.moderate_nat is not None else None)

        return AgentFeatures(
            netshield_level=features.netshield,
            randomized_nat=randomized_nat,
            split_tcp=features.vpn_accelerator,
            port_forwarding=features.port_forwarding,
            bouncing=self._vpnserver.label,
            jail=None  # Currently not in use
        )



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
import random
from enum import IntFlag
from typing import List, Dict

from proton.vpn.session.exceptions import ServerNotFoundError
from proton.vpn.session.servers.country_codes import get_country_name_by_code


class TierEnum(IntFlag):
    """Contains the tiers used throughout the clients.
        The tier either block or unblock certain features and/or servers/countries.
    """
    FREE = 0
    PLUS = 2
    PM = 3  # "implicit-flag-alias" has been added in 2.17.5, anything lower will throw an error.


class ServerFeatureEnum(IntFlag):
    """
    A Class representing the Server features as encoded in the feature flags field of the API:
    """
    SECURE_CORE = 1 << 0  # 1
    TOR = 1 << 1  # 2
    P2P = 1 << 2  # 4
    STREAMING = 1 << 3  # 8
    IPV6 = 1 << 4  # 16


class PhysicalServer:
    """
    A physical server instance contains the network information
    to initiate a VPN connection to the server.
    """

    def __init__(self, data: Dict):
        self._data = data

    @property
    def id(self) -> str:  # pylint: disable=invalid-name
        """Returns the physical ID of the server."""
        return self._data.get("ID")

    @property
    def entry_ip(self) -> str:
        """Returns the IP of the entered server."""
        return self._data.get("EntryIP")

    @property
    def exit_ip(self) -> str:
        """Returns the IP of the exited server.
            If you want to display to which IP a user is connected
            then use this one.
        """
        return self._data.get("ExitIP")

    @property
    def domain(self) -> str:
        """Returns the Domain of the connected server.
            This is usually used for TLS Authentication.
        """
        return self._data.get("Domain")

    @property
    def enabled(self) -> bool:
        """Returns if the server is enabled or not"""
        return self._data.get("Status") == 1

    @property
    def generation(self) -> str:
        """Returns the generation of the server."""
        return self._data.get("Generation")

    @property
    def label(self) -> str:
        """Returns the label value.
            If label is passed then it ensures that the
            `ExitIP` matches exactly to the server that we're connected.
        """
        return self._data.get("Label")

    @property
    def services_down_reason(self) -> str:
        """Returns the reason of why the servers are down."""
        return self._data.get("ServicesDownReason")

    @property
    def x25519_pk(self) -> str:
        """ X25519 public key of the physical available as a base64 encoded string.
        """
        return self._data.get("X25519PublicKey")

    def __repr__(self):
        if self.label != '':
            return f'PhysicalServer<{self.domain}+b:{self.label}>'

        return f'PhysicalServer<{self.domain}>'


class LogicalServer:  # pylint: disable=too-many-public-methods
    """
    Abstraction of a VPN server.

    One logical servers abstract one or more
    PhysicalServer instances away.
    """

    def __init__(self, data: Dict):
        self._data = data

    def update(self, server_load: ServerLoad):
        """Internally updates the logical server:
            * Load
            * Score
            * Status
        """
        if self.id != server_load.id:
            raise ValueError(
                "The id of the logical server does not match the one of "
                "the server load object"
            )

        self._data["Load"] = server_load.load
        self._data["Score"] = server_load.score
        self._data["Status"] = 1 if server_load.enabled else 0

    @property
    def id(self) -> str:  # pylint: disable=invalid-name
        """Returns the id of the logical server."""
        return self._data.get("ID")

    # Score, load and status can be modified (needed to update loads)
    @property
    def load(self) -> int:
        """Returns the load of the servers.
            This is generally only used for UI purposes.
        """
        return self._data.get("Load")

    @property
    def score(self) -> float:
        """Returns the score of the server.
            The score is automatically calculated by the API and
            is used for the logic of the "Quick Connect".
            The lower the number is the better is for establishing a connection.
        """
        return self._data.get("Score")

    @property
    def enabled(self) -> bool:
        """Returns if the server is enabled or not.
            Usually the API should return 0 if all physical servers
            are not enabled, but just to be sure we also evaluate all
            physical servers.
        """
        return self._data.get("Status") == 1 and any(
            x.enabled for x in self.physical_servers
        )

    @property
    def under_maintenance(self) -> bool:
        """Returns whether this server should be considered under maintenance or not."""
        return not self.enabled

    # Every other propriety is readonly
    @property
    def name(self) -> str:
        """Name of the logical, ie: CH#10"""
        return self._data.get("Name")

    @property
    def entry_country(self) -> str:
        """2 letter country code entry, ie: CH"""
        return self._data.get("EntryCountry")

    @property
    def entry_country_name(self) -> str:
        """Full name of the entry country (e.g. Switzerland)."""
        return get_country_name_by_code(self.entry_country)

    @property
    def exit_country(self) -> str:
        """2 letter country code exit, ie: CH"""
        return self._data.get("ExitCountry")

    @property
    def exit_country_name(self) -> str:
        """Full name of the exit country (e.g. Argentina)."""
        return get_country_name_by_code(self.exit_country)

    @property
    def host_country(self) -> str:
        """2 letter country code host: CH.
            If there is a host country then it means that this server location
            is emulated, see Smart Routing definition for further clarification.
        """
        return self._data.get("HostCountry")

    @property
    def smart_routing(self) -> bool:
        """Returns whether this server uses smart routing technology or not."""
        return self.host_country is not None

    @property
    def features(self) -> List[ServerFeatureEnum]:
        """ List of features supported by this Logical."""
        return self.__unpack_bitmap_features(self._data.get("Features", 0))

    def __unpack_bitmap_features(self, server_value):
        server_features = [
            feature_enum
            for feature_enum
            in ServerFeatureEnum
            if (server_value & feature_enum) != 0
        ]
        return server_features

    @property
    def region(self) -> str:
        """Returns the region of the server."""
        return self._data.get("Region")

    @property
    def city(self) -> str:
        """Returns the city of the server."""
        return self._data.get("City")

    @property
    def tier(self) -> int:
        """Returns the minimum required tier to be able to establish a connection.
            Server-side check is always done, so this is mainly for UI purposes.
        """
        return TierEnum(int(self._data.get("Tier")))

    @property
    def free(self) -> bool:
        """Returns whether this server is available to the free tier or not."""
        return self.tier == TierEnum.FREE

    @property
    def latitude(self) -> float:
        """Returns servers latitude."""
        return self._data.get("Location", {}).get("Lat")

    @property
    def longitude(self) -> float:
        """Returns servers longitude."""
        return self._data.get("Location", {}).get("Long")

    @property
    def data(self) -> dict:
        """Returns a copy of the data pertaining this server."""
        return self._data.copy()

    @property
    def physical_servers(self) -> List[PhysicalServer]:
        """ Get all the physicals of supporting a logical
        """
        return [PhysicalServer(x) for x in self._data.get("Servers", [])]

    def get_random_physical_server(self) -> PhysicalServer:
        """ Get a random `enabled` physical linked to this logical
        """
        enabled_servers = [x for x in self.physical_servers if x.enabled]
        if len(enabled_servers) == 0:
            raise ServerNotFoundError("No physical servers could be found")

        return random.choice(enabled_servers)  # nosec B311 # noqa: E501 # pylint: disable=line-too-long # nosemgrep: gitlab.bandit.B311

    def to_dict(self) -> Dict:
        """Converts this object to a dictionary for serialization purposes."""
        return self._data

    def __repr__(self):
        return f'LogicalServer<{self._data.get("Name", "??")}>'


class ServerLoad:
    """Contains data about logical servers to be updated frequently.
    """

    def __init__(self, data: Dict):
        self._data = data

    @property
    def id(self) -> str:  # pylint: disable=invalid-name
        """Returns the id of the logical server."""
        return self._data.get("ID")

    @property
    def load(self) -> int:
        """Returns the load of the servers.
            This is generally only used for UI purposes.
        """
        return self._data.get("Load")

    @property
    def score(self) -> float:
        """Returns the score of the server.
            The score is automatically calculated by the API and
            is used for the logic of the "Quick Connect".
            The lower the number is the better is for establishing a connection.
        """
        return self._data.get("Score")

    @property
    def enabled(self) -> bool:
        """Returns if the server is enabled or not.
        """
        return self._data.get("Status") == 1

    def __str__(self):
        return str(self._data)



"""
Copyright (c) 2025 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from importlib.metadata import version
from packaging.version import Version

fido2_version = Version(version("fido2"))
if fido2_version >= Version("2.0.0"):
    from proton.vpn.session.fido2_2 import (create_client,  # noqa: F401
                                            create_options,
                                            create_from_client_assertion)
elif fido2_version >= Version("1.1.2"):
    from proton.vpn.session.fido2_1 import (create_client,  # noqa: F401
                                            create_options,
                                            create_from_client_assertion)
else:
    raise ImportError(
        f"python3-fido2 version {fido2_version} not supported. "
        "Version 1.1.2 or higher required."
    )

__all__ = [
    "create_client",
    "create_options",
    "create_from_client_assertion"
]



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import base64
import hashlib
from typing import Optional

import cryptography.hazmat.primitives.asymmetric
from cryptography.hazmat.primitives.serialization import Encoding, PrivateFormat
from cryptography.hazmat.primitives import serialization
import nacl.bindings


class KeyHandler:  # pylint: disable=missing-class-docstring

    PREFIX_SK = bytes(
        [int(x, 16) for x in '30:2E:02:01:00:30:05:06:03:2B:65:70:04:22:04:20'.split(':')]
    )
    PREFIX_PK = bytes([int(x, 16) for x in '30:2A:30:05:06:03:2B:65:70:03:21:00'.split(':')])

    def __init__(self, private_key=None):
        """ private key parameter must be in ed25519 format,
            from which we convert to x25519 format with nacl.
            But it's not possible to convert from x25519 to ed25519.
        """
        self._private_key, self._public_key = self.__generate_key_pair(private_key=private_key)
        tmp_ed25519_sk = self.ed25519_sk_bytes
        tmp_ed25519_pk = self.ed25519_pk_bytes
        """
        # crypto_sign_ed25519_sk_to_curve25519() is equivalent to :
        tmp = list(hashlib.sha512(ed25519_sk).digest()[:32])
        tmp[0] &= 248
        tmp[31] &= 127
        tmp[31] |= 64
        self._x25519_sk = bytes(tmp)
        """
        self._x25519_sk = nacl.bindings.crypto_sign_ed25519_sk_to_curve25519(
            tmp_ed25519_sk + tmp_ed25519_pk
        )
        self._x25519_pk = nacl.bindings.crypto_sign_ed25519_pk_to_curve25519(tmp_ed25519_pk)

    @classmethod
    def get_proton_fingerprint_from_x25519_pk(cls, x25519_pk: bytes) -> str:  # noqa: E501 pylint: disable=missing-function-docstring
        return base64.b64encode(hashlib.sha512(x25519_pk).digest()).decode("ascii")

    @classmethod
    def from_sk_file(cls, ed25519sk_file):  # pylint: disable=missing-function-docstring
        backend_default = None
        # cryptography.sys.version_info not available in 2.6
        crypto_major, crypto_minor = cryptography.__version__.split(".")[:2]

        if int(crypto_major) < 3 or \
           int(crypto_major) == 3 and \
           int(crypto_minor) < 1:

            # backend is required if library < 3.1
            backend_default = cryptography.hazmat.backends.default_backend()

        with open(file=ed25519sk_file) as file:  # pylint: disable=unspecified-encoding
            pem_data = "".join(file.readlines())

        key = serialization.load_pem_private_key(
            pem_data.encode("ascii"), password=None, backend=backend_default
        )

        assert isinstance(  # nosec B311, B101 # noqa: E501 # pylint: disable=line-too-long # nosemgrep: gitlab.bandit.B101
            key,
            cryptography.hazmat.primitives.asymmetric.ed25519.Ed25519PrivateKey)  # nosec: B101
        private_key = key.private_bytes(
            Encoding.Raw, PrivateFormat.Raw,
            encryption_algorithm=serialization.NoEncryption()
        )
        return KeyHandler(private_key=private_key)

    @property
    def ed25519_sk_str(self) -> str:  # pylint: disable=missing-function-docstring
        return base64.b64encode(self.ed25519_sk_bytes).decode("ascii")

    @property
    def ed25519_sk_bytes(self) -> bytes:  # pylint: disable=missing-function-docstring
        return self._private_key.private_bytes(
            Encoding.Raw, PrivateFormat.Raw,
            encryption_algorithm=serialization.NoEncryption()
        )

    @property
    def ed25519_pk_bytes(self) -> bytes:  # pylint: disable=missing-function-docstring
        return self._public_key.public_bytes(
            encoding=serialization.Encoding.Raw,
            format=serialization.PublicFormat.Raw
        )

    @property
    def ed25519_pk_str_asn1(self) -> bytes:  # pylint: disable=missing-function-docstring
        return base64.b64encode(self.PREFIX_PK + self.ed25519_pk_bytes)

    @property
    def ed25519_sk_pem(self) -> str:  # pylint: disable=missing-function-docstring
        return self.get_ed25519_sk_pem()

    @property
    def ed25519_pk_pem(self) -> str:  # pylint: disable=missing-function-docstring
        return self._public_key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo
        ).decode('ascii')

    @property
    def x25519_sk_bytes(self) -> bytes:  # pylint: disable=missing-function-docstring
        return self._x25519_sk

    @property
    def x25519_pk_bytes(self) -> bytes:  # pylint: disable=missing-function-docstring
        return self._x25519_pk

    @property
    def x25519_sk_str(self) -> str:  # pylint: disable=missing-function-docstring
        return base64.b64encode(self._x25519_sk).decode("ascii")

    @property
    def x25519_pk_str(self) -> str:  # pylint: disable=missing-function-docstring
        return base64.b64encode(self._x25519_pk).decode("ascii")

    @classmethod
    def __generate_key_pair(cls, private_key=None):
        if private_key:
            private_key = cryptography.hazmat.primitives.asymmetric\
                .ed25519.Ed25519PrivateKey.from_private_bytes(private_key)
        else:
            private_key = cryptography.hazmat.primitives.asymmetric\
                .ed25519.Ed25519PrivateKey.generate()
        public_key = private_key.public_key()
        return private_key, public_key

    def get_ed25519_sk_pem(self, password: Optional[bytes] = None) -> str:
        """
        Returns the ed5519 private key in pem format,
        and encrypted if a password was passed.
        """
        if password:
            encryption_algorithm = serialization.BestAvailableEncryption(password=password)
        else:
            encryption_algorithm = serialization.NoEncryption()

        return self._private_key.private_bytes(
            encoding=Encoding.PEM, format=PrivateFormat.PKCS8,
            encryption_algorithm=encryption_algorithm
        ).decode('ascii')


def bytes_to_str_hexa(b: bytes):  # pylint: disable=missing-function-docstring invalid-name
    return ":".join(["{:02x}".format(x) for x in b])  # pylint: disable=consider-using-f-string
